import sys
from unittest.mock import ANY, call, MagicMock, patch

import pytest
import torch

from ignite.engine import Engine, Events, State

from ignite.handlers.visdom_logger import (
    _DummyExecutor,
    global_step_from_engine,
    GradsScalarHandler,
    OptimizerParamsHandler,
    OutputHandler,
    VisdomLogger,
    WeightsScalarHandler,
)

# Run tests on a single worker to avoid issues with connecting to the visdom
# server This requires that the --dist=loadgroup option be passed to pytest.
pytestmark = [pytest.mark.timeout(30), pytest.mark.xdist_group(name="visdom")]


def test_optimizer_params_handler_wrong_setup():
    with pytest.raises(TypeError):
        OptimizerParamsHandler(optimizer=None)

    optimizer = MagicMock(spec=torch.optim.Optimizer)
    handler = OptimizerParamsHandler(optimizer=optimizer)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler OptimizerParamsHandler works only with VisdomLogger"):
        handler(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_optimizer_params():
    optimizer = torch.optim.SGD([torch.tensor(0.0)], lr=0.01)
    wrapper = OptimizerParamsHandler(optimizer=optimizer, param_name="lr")
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    # mock_logger.vis.line.assert_called_once_with("lr/group_0", 0.01, 123)
    assert len(wrapper.windows) == 1 and "lr/group_0" in wrapper.windows
    assert wrapper.windows["lr/group_0"]["win"] is not None

    mock_logger.vis.line.assert_called_once_with(
        X=[123],
        Y=[0.01],
        env=mock_logger.vis.env,
        win=None,
        update=None,
        opts=wrapper.windows["lr/group_0"]["opts"],
        name="lr/group_0",
    )

    wrapper = OptimizerParamsHandler(optimizer=optimizer, param_name="lr", tag="generator")
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 1 and "generator/lr/group_0" in wrapper.windows
    assert wrapper.windows["generator/lr/group_0"]["win"] is not None

    mock_logger.vis.line.assert_called_once_with(
        X=[123],
        Y=[0.01],
        env=mock_logger.vis.env,
        win=None,
        update=None,
        opts=wrapper.windows["generator/lr/group_0"]["opts"],
        name="generator/lr/group_0",
    )


def test_output_handler_with_wrong_logger_type():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler 'OutputHandler' works only with VisdomLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_output_handler_output_transform(dirname):
    wrapper = OutputHandler("tag", output_transform=lambda x: x)
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.output = 12345
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 1 and "tag/output" in wrapper.windows
    assert wrapper.windows["tag/output"]["win"] is not None

    mock_logger.vis.line.assert_called_once_with(
        X=[123],
        Y=[12345],
        env=mock_logger.vis.env,
        win=None,
        update=None,
        opts=wrapper.windows["tag/output"]["opts"],
        name="tag/output",
    )

    wrapper = OutputHandler("another_tag", output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 1 and "another_tag/loss" in wrapper.windows
    assert wrapper.windows["another_tag/loss"]["win"] is not None

    mock_logger.vis.line.assert_called_once_with(
        X=[123],
        Y=[12345],
        env=mock_logger.vis.env,
        win=None,
        update=None,
        opts=wrapper.windows["another_tag/loss"]["opts"],
        name="another_tag/loss",
    )


def test_output_handler_metric_names(dirname):
    wrapper = OutputHandler("tag", metric_names=["a", "b"])
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 2 and "tag/a" in wrapper.windows and "tag/b" in wrapper.windows
    assert wrapper.windows["tag/a"]["win"] is not None
    assert wrapper.windows["tag/b"]["win"] is not None

    assert mock_logger.vis.line.call_count == 2
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[12.23],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a"]["opts"],
                name="tag/a",
            ),
            call(
                X=[5],
                Y=[23.45],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/b"]["opts"],
                name="tag/b",
            ),
        ],
        any_order=True,
    )

    wrapper = OutputHandler("tag", metric_names=["a"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": torch.tensor([0.0, 1.0, 2.0, 3.0])})
    mock_engine.state.iteration = 5

    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 4 and all([f"tag/a/{i}" in wrapper.windows for i in range(4)])
    assert wrapper.windows["tag/a/0"]["win"] is not None
    assert wrapper.windows["tag/a/1"]["win"] is not None
    assert wrapper.windows["tag/a/2"]["win"] is not None
    assert wrapper.windows["tag/a/3"]["win"] is not None

    assert mock_logger.vis.line.call_count == 4
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[0.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a/0"]["opts"],
                name="tag/a/0",
            ),
            call(
                X=[5],
                Y=[1.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a/1"]["opts"],
                name="tag/a/1",
            ),
            call(
                X=[5],
                Y=[2.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a/2"]["opts"],
                name="tag/a/2",
            ),
            call(
                X=[5],
                Y=[3.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a/3"]["opts"],
                name="tag/a/3",
            ),
        ],
        any_order=True,
    )

    wrapper = OutputHandler("tag", metric_names=["a", "c"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 55.56, "c": "Some text"})
    mock_engine.state.iteration = 7

    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    with pytest.warns(UserWarning):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 1 and "tag/a" in wrapper.windows
    assert wrapper.windows["tag/a"]["win"] is not None

    assert mock_logger.vis.line.call_count == 1
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[7],
                Y=[55.56],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a"]["opts"],
                name="tag/a",
            ),
        ],
        any_order=True,
    )

    # all metrics
    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert len(wrapper.windows) == 2 and "tag/a" in wrapper.windows and "tag/b" in wrapper.windows
    assert wrapper.windows["tag/a"]["win"] is not None
    assert wrapper.windows["tag/b"]["win"] is not None

    assert mock_logger.vis.line.call_count == 2
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[12.23],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a"]["opts"],
                name="tag/a",
            ),
            call(
                X=[5],
                Y=[23.45],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/b"]["opts"],
                name="tag/b",
            ),
        ],
        any_order=True,
    )

    # all metrics
    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State(
        metrics={
            "a": 123,
            "b": {"c": [2.34, {"d": 1}]},
            "c": (22, [33, -5.5], {"e": 32.1}),
        }
    )
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    expected_metrics = {
        "tag/a": 123,
        "tag/b/c/0": 2.34,
        "tag/b/c/1/d": 1,
        "tag/c/0": 22,
        "tag/c/1/0": 33,
        "tag/c/1/1": -5.5,
        "tag/c/2/e": 32.1,
    }

    assert len(wrapper.windows) == len(expected_metrics)
    for key in expected_metrics:
        assert key in wrapper.windows
        assert wrapper.windows[key]["win"] is not None
    assert mock_logger.vis.line.call_count == len(expected_metrics)
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[value],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows[key]["opts"],
                name=key,
            )
            for key, value in expected_metrics.items()
        ],
        any_order=True,
    )


def test_output_handler_both(dirname):
    wrapper = OutputHandler("tag", metric_names=["a", "b"], output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    assert mock_logger.vis.line.call_count == 3
    assert (
        len(wrapper.windows) == 3
        and "tag/a" in wrapper.windows
        and "tag/b" in wrapper.windows
        and "tag/loss" in wrapper.windows
    )
    assert wrapper.windows["tag/a"]["win"] is not None
    assert wrapper.windows["tag/b"]["win"] is not None
    assert wrapper.windows["tag/loss"]["win"] is not None

    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[12.23],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/a"]["opts"],
                name="tag/a",
            ),
            call(
                X=[5],
                Y=[23.45],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/b"]["opts"],
                name="tag/b",
            ),
            call(
                X=[5],
                Y=[12345],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/loss"]["opts"],
                name="tag/loss",
            ),
        ],
        any_order=True,
    )

    mock_engine.state.epoch = 6
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    assert mock_logger.vis.line.call_count == 6
    assert (
        len(wrapper.windows) == 3
        and "tag/a" in wrapper.windows
        and "tag/b" in wrapper.windows
        and "tag/loss" in wrapper.windows
    )
    assert wrapper.windows["tag/a"]["win"] is not None
    assert wrapper.windows["tag/b"]["win"] is not None
    assert wrapper.windows["tag/loss"]["win"] is not None

    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[6],
                Y=[12.23],
                env=mock_logger.vis.env,
                win=wrapper.windows["tag/a"]["win"],
                update="append",
                opts=wrapper.windows["tag/a"]["opts"],
                name="tag/a",
            ),
            call(
                X=[6],
                Y=[23.45],
                env=mock_logger.vis.env,
                win=wrapper.windows["tag/b"]["win"],
                update="append",
                opts=wrapper.windows["tag/b"]["opts"],
                name="tag/b",
            ),
            call(
                X=[6],
                Y=[12345],
                env=mock_logger.vis.env,
                win=wrapper.windows["tag/loss"]["win"],
                update="append",
                opts=wrapper.windows["tag/loss"]["opts"],
                name="tag/loss",
            ),
        ],
        any_order=True,
    )


def test_output_handler_state_attrs():
    wrapper = OutputHandler("tag", state_attributes=["alpha", "beta", "gamma"])
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 5
    mock_engine.state.alpha = 3.899
    mock_engine.state.beta = torch.tensor(12.0)
    mock_engine.state.gamma = torch.tensor([21.0, 6.0])

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.vis.line.call_count == 4
    assert (
        len(wrapper.windows) == 4
        and "tag/alpha" in wrapper.windows
        and "tag/beta" in wrapper.windows
        and "tag/gamma/0" in wrapper.windows
        and "tag/gamma/1" in wrapper.windows
    )
    assert wrapper.windows["tag/alpha"]["win"] is not None
    assert wrapper.windows["tag/beta"]["win"] is not None
    assert wrapper.windows["tag/gamma/0"]["win"] is not None
    assert wrapper.windows["tag/gamma/1"]["win"] is not None

    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[3.899],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/alpha"]["opts"],
                name="tag/alpha",
            ),
            call(
                X=[5],
                Y=[12.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/beta"]["opts"],
                name="tag/beta",
            ),
            call(
                X=[5],
                Y=[21.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/gamma/0"]["opts"],
                name="tag/gamma/0",
            ),
            call(
                X=[5],
                Y=[6.0],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/gamma/1"]["opts"],
                name="tag/gamma/1",
            ),
        ],
        any_order=True,
    )


def test_output_handler_with_wrong_global_step_transform_output():
    def global_step_transform(*args, **kwargs):
        return "a"

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    with pytest.raises(TypeError, match="global_step must be int"):
        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)


def test_output_handler_with_global_step_transform():
    def global_step_transform(*args, **kwargs):
        return 10

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.vis.line.call_count == 1
    assert len(wrapper.windows) == 1 and "tag/loss" in wrapper.windows
    assert wrapper.windows["tag/loss"]["win"] is not None

    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[10],
                Y=[12345],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/loss"]["opts"],
                name="tag/loss",
            )
        ]
    )


def test_output_handler_with_global_step_from_engine():
    mock_another_engine = MagicMock()
    mock_another_engine.state = State()
    mock_another_engine.state.epoch = 10
    mock_another_engine.state.output = 12.345

    wrapper = OutputHandler(
        "tag",
        output_transform=lambda x: {"loss": x},
        global_step_transform=global_step_from_engine(mock_another_engine),
    )

    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 1
    mock_engine.state.output = 0.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.vis.line.call_count == 1
    assert len(wrapper.windows) == 1 and "tag/loss" in wrapper.windows
    assert wrapper.windows["tag/loss"]["win"] is not None
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[mock_another_engine.state.epoch],
                Y=[mock_engine.state.output],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["tag/loss"]["opts"],
                name="tag/loss",
            )
        ]
    )

    mock_another_engine.state.epoch = 11
    mock_engine.state.output = 1.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.vis.line.call_count == 2
    assert len(wrapper.windows) == 1 and "tag/loss" in wrapper.windows
    assert wrapper.windows["tag/loss"]["win"] is not None
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[mock_another_engine.state.epoch],
                Y=[mock_engine.state.output],
                env=mock_logger.vis.env,
                win=wrapper.windows["tag/loss"]["win"],
                update="append",
                opts=wrapper.windows["tag/loss"]["opts"],
                name="tag/loss",
            )
        ]
    )


def test_weights_scalar_handler_wrong_setup():
    with pytest.raises(TypeError, match="Argument model should be of type torch.nn.Module"):
        WeightsScalarHandler(None)

    model = MagicMock(spec=torch.nn.Module)
    with pytest.raises(TypeError, match="Argument reduction should be callable"):
        WeightsScalarHandler(model, reduction=123)

    with pytest.raises(TypeError, match="Output of the reduction function should be a scalar"):
        WeightsScalarHandler(model, reduction=lambda x: x)

    wrapper = WeightsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler 'WeightsScalarHandler' works only with VisdomLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_weights_scalar_handler():
    class DummyModel(torch.nn.Module):
        def __init__(self):
            super(DummyModel, self).__init__()
            self.fc1 = torch.nn.Linear(10, 10)
            self.fc2 = torch.nn.Linear(12, 12)
            self.fc1.weight.data.zero_()
            self.fc1.bias.data.zero_()
            self.fc2.weight.data.fill_(1.0)
            self.fc2.bias.data.fill_(1.0)

    model = DummyModel()

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = WeightsScalarHandler(model, tag=tag)
        mock_logger = MagicMock(spec=VisdomLogger)
        mock_logger.vis = MagicMock()
        mock_logger.executor = _DummyExecutor()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert mock_logger.vis.line.call_count == 4
        mock_logger.vis.line.assert_has_calls(
            [
                call(
                    X=[5],
                    Y=[0.0],
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "weights_norm/fc1/weight"]["opts"],
                    name=tag_prefix + "weights_norm/fc1/weight",
                ),
                call(
                    X=[5],
                    Y=[0.0],
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "weights_norm/fc1/bias"]["opts"],
                    name=tag_prefix + "weights_norm/fc1/bias",
                ),
                call(
                    X=[5],
                    Y=[12.0],
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "weights_norm/fc2/weight"]["opts"],
                    name=tag_prefix + "weights_norm/fc2/weight",
                ),
                call(
                    X=[5],
                    Y=ANY,
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "weights_norm/fc2/bias"]["opts"],
                    name=tag_prefix + "weights_norm/fc2/bias",
                ),
            ],
            any_order=True,
        )

    _test()
    _test(tag="tag")


def test_weights_scalar_handler_custom_reduction():
    class DummyModel(torch.nn.Module):
        def __init__(self):
            super(DummyModel, self).__init__()
            self.fc1 = torch.nn.Linear(10, 10)
            self.fc2 = torch.nn.Linear(12, 12)
            self.fc1.weight.data.zero_()
            self.fc1.bias.data.zero_()
            self.fc2.weight.data.fill_(1.0)
            self.fc2.bias.data.fill_(1.0)

    model = DummyModel()

    def norm(x):
        return 12.34

    wrapper = WeightsScalarHandler(model, reduction=norm, show_legend=True)
    mock_logger = MagicMock(spec=VisdomLogger)
    mock_logger.vis = MagicMock()
    mock_logger.executor = _DummyExecutor()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    assert mock_logger.vis.line.call_count == 4
    mock_logger.vis.line.assert_has_calls(
        [
            call(
                X=[5],
                Y=[12.34],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["weights_norm/fc1/weight"]["opts"],
                name="weights_norm/fc1/weight",
            ),
            call(
                X=[5],
                Y=[12.34],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["weights_norm/fc1/bias"]["opts"],
                name="weights_norm/fc1/bias",
            ),
            call(
                X=[5],
                Y=[12.34],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["weights_norm/fc2/weight"]["opts"],
                name="weights_norm/fc2/weight",
            ),
            call(
                X=[5],
                Y=[12.34],
                env=mock_logger.vis.env,
                win=None,
                update=None,
                opts=wrapper.windows["weights_norm/fc2/bias"]["opts"],
                name="weights_norm/fc2/bias",
            ),
        ],
        any_order=True,
    )


def test_grads_scalar_handler_wrong_setup():
    with pytest.raises(TypeError, match="Argument model should be of type torch.nn.Module"):
        GradsScalarHandler(None)

    model = MagicMock(spec=torch.nn.Module)
    with pytest.raises(TypeError, match="Argument reduction should be callable"):
        GradsScalarHandler(model, reduction=123)

    wrapper = GradsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler 'GradsScalarHandler' works only with VisdomLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_grads_scalar_handler(dummy_model_factory, norm_mock):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = GradsScalarHandler(model, reduction=norm_mock, tag=tag)
        mock_logger = MagicMock(spec=VisdomLogger)
        mock_logger.vis = MagicMock()
        mock_logger.executor = _DummyExecutor()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert mock_logger.vis.line.call_count == 4
        mock_logger.vis.line.assert_has_calls(
            [
                call(
                    X=[5],
                    Y=ANY,
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "grads_norm/fc1/weight"]["opts"],
                    name=tag_prefix + "grads_norm/fc1/weight",
                ),
                call(
                    X=[5],
                    Y=ANY,
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "grads_norm/fc1/bias"]["opts"],
                    name=tag_prefix + "grads_norm/fc1/bias",
                ),
                call(
                    X=[5],
                    Y=ANY,
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "grads_norm/fc2/weight"]["opts"],
                    name=tag_prefix + "grads_norm/fc2/weight",
                ),
                call(
                    X=[5],
                    Y=ANY,
                    env=mock_logger.vis.env,
                    win=None,
                    update=None,
                    opts=wrapper.windows[tag_prefix + "grads_norm/fc2/bias"]["opts"],
                    name=tag_prefix + "grads_norm/fc2/bias",
                ),
            ],
            any_order=True,
        )

    _test()
    _test(tag="tag")


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration_no_server():
    with pytest.raises(ConnectionError, match="Error connecting to Visdom server"):
        VisdomLogger()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_logger_init_hostname_port(visdom_server):
    # Explicit hostname, port
    vd_logger = VisdomLogger(server=visdom_server[0], port=visdom_server[1], num_workers=0)
    assert "main" in vd_logger.vis.get_env_list()
    vd_logger.close()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_logger_init_env_vars(visdom_server):
    # As env vars
    import os

    os.environ["VISDOM_SERVER_URL"] = visdom_server[0]
    os.environ["VISDOM_PORT"] = str(visdom_server[1])
    vd_logger = VisdomLogger(server=visdom_server[0], port=visdom_server[1], num_workers=0)
    assert "main" in vd_logger.vis.get_env_list()
    vd_logger.close()


def _parse_content(content):
    import json

    return json.loads(content)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration_no_executor(visdom_server):
    vd_logger = VisdomLogger(server=visdom_server[0], port=visdom_server[1], num_workers=0)

    # close all windows in 'main' environment
    vd_logger.vis.close()

    n_epochs = 3
    data = list(range(10))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    trainer = Engine(update_fn)
    output_handler = OutputHandler(tag="training", output_transform=lambda x: {"loss": x})
    vd_logger.attach(trainer, log_handler=output_handler, event_name=Events.ITERATION_COMPLETED)

    trainer.run(data, max_epochs=n_epochs)

    assert len(output_handler.windows) == 1
    assert "training/loss" in output_handler.windows
    win_name = output_handler.windows["training/loss"]["win"]
    data = vd_logger.vis.get_window_data(win=win_name)
    data = _parse_content(data)
    assert "content" in data and "data" in data["content"]
    data = data["content"]["data"][0]
    assert "x" in data and "y" in data
    x_vals, y_vals = data["x"], data["y"]
    assert all([int(x) == x_true for x, x_true in zip(x_vals, list(range(1, n_epochs * len(data) + 1)))])
    assert all([y == y_true for y, y_true in zip(y_vals, losses)])
    vd_logger.close()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration_with_executor(visdom_server):
    vd_logger = VisdomLogger(server=visdom_server[0], port=visdom_server[1], num_workers=1)

    # close all windows in 'main' environment
    vd_logger.vis.close()

    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    trainer = Engine(update_fn)
    output_handler = OutputHandler(tag="training", output_transform=lambda x: {"loss": x})
    vd_logger.attach(trainer, log_handler=output_handler, event_name=Events.ITERATION_COMPLETED)

    trainer.run(data, max_epochs=n_epochs)

    assert len(output_handler.windows) == 1
    assert "training/loss" in output_handler.windows
    win_name = output_handler.windows["training/loss"]["win"]
    data = vd_logger.vis.get_window_data(win=win_name)
    data = _parse_content(data)
    assert "content" in data and "data" in data["content"]
    data = data["content"]["data"][0]
    assert "x" in data and "y" in data
    x_vals, y_vals = data["x"], data["y"]
    assert all([int(x) == x_true for x, x_true in zip(x_vals, list(range(1, n_epochs * len(data) + 1)))])
    assert all([y == y_true for y, y_true in zip(y_vals, losses)])

    vd_logger.close()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration_with_executor_as_context_manager(visdom_server):
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    with VisdomLogger(server=visdom_server[0], port=visdom_server[1], num_workers=1) as vd_logger:
        # close all windows in 'main' environment
        vd_logger.vis.close()

        trainer = Engine(update_fn)
        output_handler = OutputHandler(tag="training", output_transform=lambda x: {"loss": x})
        vd_logger.attach(trainer, log_handler=output_handler, event_name=Events.ITERATION_COMPLETED)

        trainer.run(data, max_epochs=n_epochs)

        assert len(output_handler.windows) == 1
        assert "training/loss" in output_handler.windows
        win_name = output_handler.windows["training/loss"]["win"]
        data = vd_logger.vis.get_window_data(win=win_name)
        data = _parse_content(data)
        assert "content" in data and "data" in data["content"]
        data = data["content"]["data"][0]
        assert "x" in data and "y" in data
        x_vals, y_vals = data["x"], data["y"]
        assert all([int(x) == x_true for x, x_true in zip(x_vals, list(range(1, n_epochs * len(data) + 1)))])
        assert all([y == y_true for y, y_true in zip(y_vals, losses)])


@pytest.mark.parametrize("no_site_packages", ["visdom"], indirect=True)
def test_no_visdom(no_site_packages):
    with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires visdom package"):
        VisdomLogger()


def test_no_concurrent():
    with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires concurrent.futures"):
        with patch.dict("sys.modules", {"concurrent.futures": None}):
            VisdomLogger(num_workers=1)



import warnings
from functools import partial
from itertools import accumulate

import numpy as np
import pytest
import torch

import ignite.distributed as idist
from ignite.engine import Engine, Events
from ignite.metrics import Accuracy, RunningAverage
from ignite.metrics.metric import RunningBatchWise, RunningEpochWise, SingleEpochRunningBatchWise


def test_wrong_input_args():
    with pytest.raises(TypeError, match=r"Argument src should be a Metric or None."):
        RunningAverage(src=[12, 34])

    with pytest.raises(ValueError, match=r"Argument alpha should be a float between"):
        RunningAverage(alpha=-1.0)

    with pytest.raises(ValueError, match=r"Argument output_transform should be None if src is a Metric"):
        RunningAverage(Accuracy(), output_transform=lambda x: x[0])

    with pytest.raises(ValueError, match=r"Argument output_transform should not be None if src corresponds"):
        RunningAverage()

    with pytest.raises(ValueError, match=r"Argument device should be None if src is a Metric"):
        RunningAverage(Accuracy(), device="cpu")

    with pytest.warns(UserWarning, match=r"`epoch_bound` is deprecated and will be removed in the future."):
        m = RunningAverage(Accuracy(), epoch_bound=True)


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("epoch_bound, usage", [(False, RunningBatchWise()), (True, SingleEpochRunningBatchWise())])
def test_epoch_bound(epoch_bound, usage):
    with warnings.catch_warnings():
        metric = RunningAverage(output_transform=lambda _: _, epoch_bound=epoch_bound)
    e1 = Engine(lambda _, __: None)
    e2 = Engine(lambda _, __: None)
    metric.attach(e1, "")
    metric.epoch_bound = None
    metric.attach(e2, "", usage)
    e1._event_handlers == e2._event_handlers


@pytest.mark.parametrize("usage", [RunningBatchWise(), SingleEpochRunningBatchWise()])
def test_integration_batchwise(usage):
    torch.manual_seed(10)
    alpha = 0.98
    n_iters = 10
    batch_size = 10
    n_classes = 10
    max_epochs = 3
    data = list(range(n_iters))
    loss = torch.arange(n_iters, dtype=torch.float)
    y_true = torch.randint(0, n_classes, size=(n_iters, batch_size))
    y_pred = torch.rand(n_iters, batch_size, n_classes)

    accuracy_running_averages = torch.tensor(
        list(
            accumulate(
                map(
                    lambda y_yp: torch.sum(y_yp[1].argmax(dim=-1) == y_yp[0]).item() / y_yp[0].size(0),
                    zip(
                        y_true if isinstance(usage, SingleEpochRunningBatchWise) else y_true.repeat(max_epochs, 1),
                        y_pred if isinstance(usage, SingleEpochRunningBatchWise) else y_pred.repeat(max_epochs, 1, 1),
                    ),
                ),
                lambda ra, acc: ra * alpha + (1 - alpha) * acc,
            )
        )
    )
    if isinstance(usage, SingleEpochRunningBatchWise):
        accuracy_running_averages = accuracy_running_averages.repeat(max_epochs)

    loss_running_averages = torch.tensor(
        list(
            accumulate(
                loss if isinstance(usage, SingleEpochRunningBatchWise) else loss.repeat(max_epochs),
                lambda ra, loss_item: ra * alpha + (1 - alpha) * loss_item,
            )
        )
    )
    if isinstance(usage, SingleEpochRunningBatchWise):
        loss_running_averages = loss_running_averages.repeat(max_epochs)

    def update_fn(_, i):
        loss_value = loss[i]
        y_true_batch = y_true[i]
        y_pred_batch = y_pred[i]
        return loss_value, y_pred_batch, y_true_batch

    trainer = Engine(update_fn)

    acc_metric = RunningAverage(Accuracy(output_transform=lambda x: [x[1], x[2]]), alpha=alpha)
    acc_metric.attach(trainer, "running_avg_accuracy", usage)

    avg_output = RunningAverage(output_transform=lambda x: x[0], alpha=alpha)
    avg_output.attach(trainer, "running_avg_loss", usage)

    metric_acc_running_averages = []
    metric_loss_running_averages = []

    @trainer.on(Events.ITERATION_COMPLETED)
    def _(engine):
        metric_acc_running_averages.append(engine.state.metrics["running_avg_accuracy"])
        metric_loss_running_averages.append(engine.state.metrics["running_avg_loss"])

    trainer.run(data, max_epochs=3)

    assert (torch.tensor(metric_acc_running_averages) == accuracy_running_averages).all()
    assert (torch.tensor(metric_loss_running_averages) == loss_running_averages).all()

    metric_state = acc_metric.state_dict()
    saved__value = acc_metric._value
    saved_src__num_correct = acc_metric.src._num_correct
    saved_src__num_examples = acc_metric.src._num_examples
    acc_metric.reset()
    acc_metric.load_state_dict(metric_state)
    assert acc_metric._value == saved__value
    assert acc_metric.src._num_examples == saved_src__num_examples
    assert (acc_metric.src._num_correct == saved_src__num_correct).all()

    metric_state = avg_output.state_dict()
    saved__value = avg_output._value
    assert avg_output.src is None
    avg_output.reset()
    avg_output.load_state_dict(metric_state)
    assert avg_output._value == saved__value
    assert avg_output.src is None


def test_integration_epochwise():
    torch.manual_seed(10)
    alpha = 0.98
    n_iters = 10
    batch_size = 10
    n_classes = 10
    max_epochs = 3
    data = list(range(n_iters))
    y_true = torch.randint(0, n_classes, size=(n_iters, batch_size))
    y_pred = torch.rand(max_epochs, n_iters, batch_size, n_classes)

    accuracy_running_averages = torch.tensor(
        list(
            accumulate(
                map(
                    lambda y_pred_epoch: torch.sum(y_pred_epoch.argmax(dim=-1) == y_true).item() / y_true.numel(),
                    y_pred,
                ),
                lambda ra, acc: ra * alpha + (1 - alpha) * acc,
            )
        )
    )

    def update_fn(engine, i):
        y_true_batch = y_true[i]
        y_pred_batch = y_pred[engine.state.epoch - 1, i]
        return y_pred_batch, y_true_batch

    trainer = Engine(update_fn)

    acc_metric = RunningAverage(Accuracy(), alpha=alpha)
    acc_metric.attach(trainer, "running_avg_accuracy", RunningEpochWise())

    metric_acc_running_averages = []

    @trainer.on(Events.EPOCH_COMPLETED)
    def _(engine):
        metric_acc_running_averages.append(engine.state.metrics["running_avg_accuracy"])

    trainer.run(data, max_epochs=3)

    assert (torch.tensor(metric_acc_running_averages) == accuracy_running_averages).all()


@pytest.mark.parametrize("usage", [RunningBatchWise(), SingleEpochRunningBatchWise(), RunningEpochWise()])
def test_multiple_attach(usage):
    n_iters = 100
    errD_values = iter(np.random.rand(n_iters))
    errG_values = iter(np.random.rand(n_iters))
    D_x_values = iter(np.random.rand(n_iters))
    D_G_z1 = iter(np.random.rand(n_iters))
    D_G_z2 = iter(np.random.rand(n_iters))

    def update_fn(engine, batch):
        return {
            "errD": next(errD_values),
            "errG": next(errG_values),
            "D_x": next(D_x_values),
            "D_G_z1": next(D_G_z1),
            "D_G_z2": next(D_G_z2),
        }

    trainer = Engine(update_fn)
    alpha = 0.98

    # attach running average
    monitoring_metrics = ["errD", "errG", "D_x", "D_G_z1", "D_G_z2"]
    for metric in monitoring_metrics:
        foo = partial(lambda x, metric: x[metric], metric=metric)
        RunningAverage(alpha=alpha, output_transform=foo).attach(trainer, metric, usage)

    @trainer.on(usage.COMPLETED)
    def check_values(engine):
        values = []
        for metric in monitoring_metrics:
            values.append(engine.state.metrics[metric])

        values = set(values)
        assert len(values) == len(monitoring_metrics)

    data = list(range(n_iters))
    trainer.run(data)


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("epoch_bound", [True, False, None])
@pytest.mark.parametrize("src", [Accuracy(), None])
@pytest.mark.parametrize("usage", [RunningBatchWise(), SingleEpochRunningBatchWise(), RunningEpochWise()])
def test_detach(epoch_bound, src, usage):
    with warnings.catch_warnings():
        m = RunningAverage(src, output_transform=(lambda _: _) if src is None else None, epoch_bound=epoch_bound)
    e = Engine(lambda _, __: None)
    m.attach(e, "m", usage)
    for event_handlers in e._event_handlers.values():
        assert len(event_handlers) != 0
    m.detach(e, usage)
    for event_handlers in e._event_handlers.values():
        assert len(event_handlers) == 0


def test_output_is_tensor():
    m = RunningAverage(output_transform=lambda x: x)
    m.update(torch.rand(10, requires_grad=True).mean())
    v = m.compute()
    assert isinstance(v, torch.Tensor)
    assert not v.requires_grad

    m.update(torch.rand(10, requires_grad=True).mean())
    v = m.compute()
    assert isinstance(v, torch.Tensor)
    assert not v.requires_grad

    m.update(torch.rand(10, requires_grad=True).mean())
    v = m.compute()
    assert isinstance(v, torch.Tensor)
    assert not v.requires_grad


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    @pytest.mark.parametrize("usage", [RunningBatchWise(), SingleEpochRunningBatchWise()])
    def test_src_is_output(self, usage):
        device = idist.device()
        rank = idist.get_rank()
        n_iters = 10
        n_epochs = 3

        # Data per rank
        data = list(range(n_iters))
        rank_loss_count = n_epochs * n_iters
        all_loss_values = torch.arange(0, rank_loss_count * idist.get_world_size(), dtype=torch.float64).to(device)
        loss_values = iter(all_loss_values[rank_loss_count * rank : rank_loss_count * (rank + 1)])

        def update_fn(engine, batch):
            loss_value = next(loss_values)
            return loss_value.item()

        trainer = Engine(update_fn)
        alpha = 0.98

        metric_device = device if device.type != "xla" else "cpu"
        avg_output = RunningAverage(output_transform=lambda x: x, alpha=alpha, device=metric_device)
        avg_output.attach(trainer, "running_avg_output", usage)

        @trainer.on(usage.STARTED)
        def reset_running_avg_output(engine):
            engine.state.running_avg_output = None

        @trainer.on(usage.ITERATION_COMPLETED)
        def running_avg_output_update(engine):
            i = engine.state.iteration - 1
            o = sum([all_loss_values[i + r * rank_loss_count] for r in range(idist.get_world_size())]).item()
            o /= idist.get_world_size()
            if engine.state.running_avg_output is None:
                engine.state.running_avg_output = o
            else:
                engine.state.running_avg_output = engine.state.running_avg_output * alpha + (1.0 - alpha) * o

        @trainer.on(usage.COMPLETED)
        def assert_equal_running_avg_output_values(engine):
            it = engine.state.iteration
            assert (
                engine.state.running_avg_output == engine.state.metrics["running_avg_output"]
            ), f"{it}: {engine.state.running_avg_output} vs {engine.state.metrics['running_avg_output']}"

        trainer.run(data, max_epochs=3)

    @pytest.mark.parametrize("usage", [RunningBatchWise(), SingleEpochRunningBatchWise(), RunningEpochWise()])
    def test_src_is_metric(self, usage):
        device = idist.device()
        rank = idist.get_rank()
        n_iters = 10
        n_epochs = 3
        batch_size = 10
        n_classes = 10

        def _test(metric_device):
            data = list(range(n_iters))
            np.random.seed(12)
            all_y_true_batch_values = np.random.randint(
                0, n_classes, size=(idist.get_world_size(), n_epochs * n_iters, batch_size)
            )
            all_y_pred_batch_values = np.random.rand(idist.get_world_size(), n_epochs * n_iters, batch_size, n_classes)

            y_true_batch_values = iter(all_y_true_batch_values[rank, ...])
            y_pred_batch_values = iter(all_y_pred_batch_values[rank, ...])

            def update_fn(engine, batch):
                y_true_batch = next(y_true_batch_values)
                y_pred_batch = next(y_pred_batch_values)
                return torch.from_numpy(y_pred_batch), torch.from_numpy(y_true_batch)

            trainer = Engine(update_fn)
            alpha = 0.98

            acc_metric = RunningAverage(Accuracy(device=metric_device), alpha=alpha)
            acc_metric.attach(trainer, "running_avg_accuracy", usage)

            running_avg_acc = [
                None,
            ]
            true_acc_metric = Accuracy(device=metric_device)

            @trainer.on(Events.ITERATION_COMPLETED)
            def manual_running_avg_acc(engine):
                iteration = engine.state.iteration

                if not isinstance(usage, RunningEpochWise) or ((iteration - 1) % n_iters) == 0:
                    true_acc_metric.reset()
                if ((iteration - 1) % n_iters) == 0 and isinstance(usage, SingleEpochRunningBatchWise):
                    running_avg_acc[0] = None
                for j in range(idist.get_world_size()):
                    output = (
                        torch.from_numpy(all_y_pred_batch_values[j, iteration - 1, :, :]),
                        torch.from_numpy(all_y_true_batch_values[j, iteration - 1, :]),
                    )
                    true_acc_metric.update(output)

                if not isinstance(usage, RunningEpochWise) or (iteration % n_iters) == 0:
                    batch_acc = true_acc_metric._num_correct.item() * 1.0 / true_acc_metric._num_examples

                    if running_avg_acc[0] is None:
                        running_avg_acc[0] = batch_acc
                    else:
                        running_avg_acc[0] = running_avg_acc[0] * alpha + (1.0 - alpha) * batch_acc
                    engine.state.running_avg_acc = running_avg_acc[0]

            @trainer.on(Events.ITERATION_COMPLETED)
            def assert_equal_running_avg_acc_values(engine):
                print(engine.state.iteration)
                if not isinstance(usage, RunningEpochWise) or (
                    (engine.state.iteration > 1) and ((engine.state.iteration % n_iters) == 1)
                ):
                    assert (
                        engine.state.running_avg_acc == engine.state.metrics["running_avg_accuracy"]
                    ), f"{engine.state.running_avg_acc} vs {engine.state.metrics['running_avg_accuracy']}"

            trainer.run(data, max_epochs=3)

        _test("cpu")
        if device.type != "xla":
            _test(idist.device())

    def test_accumulator_device(self):
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(idist.device())
        for metric_device in metric_devices:
            # Don't test the src=Metric case because compute() returns a scalar,
            # so the metric doesn't accumulate on the device specified
            avg = RunningAverage(output_transform=lambda x: x, device=metric_device)
            assert avg._device == metric_device
            # Value is None until the first update then compute call

            for _ in range(3):
                avg.update(torch.tensor(1.0, device=device))
                avg.compute()

                assert (
                    avg._value.device == metric_device
                ), f"{type(avg._value.device)}:{avg._value.device} vs {type(metric_device)}:{metric_device}"



from typing import Callable, Union
from unittest.mock import MagicMock

import pytest
import torch
from sklearn.metrics import accuracy_score

import ignite.distributed as idist
from ignite.engine import Engine, State
from ignite.exceptions import NotComputableError
from ignite.metrics import Accuracy

torch.manual_seed(12)


def test_no_update():
    acc = Accuracy()
    with pytest.raises(NotComputableError, match=r"Accuracy must have at least one example before it can be computed"):
        acc.compute()


def test__check_shape():
    acc = Accuracy()

    with pytest.raises(ValueError, match=r"y and y_pred must have compatible shapes"):
        acc._check_shape((torch.randint(0, 2, size=(10, 1, 5, 12)).long(), torch.randint(0, 2, size=(10, 5, 6)).long()))

    with pytest.raises(ValueError, match=r"y and y_pred must have compatible shapes"):
        acc._check_shape((torch.randint(0, 2, size=(10, 1, 6)).long(), torch.randint(0, 2, size=(10, 5, 6)).long()))

    with pytest.raises(ValueError, match=r"y and y_pred must have compatible shapes"):
        acc._check_shape((torch.randint(0, 2, size=(10, 1)).long(), torch.randint(0, 2, size=(10, 5)).long()))


def test__check_type():
    acc = Accuracy()

    with pytest.raises(RuntimeError, match=r"Invalid shapes of y"):
        acc._check_type((torch.rand([1, 1, 1]), torch.rand([1])))


def test_binary_wrong_inputs():
    acc = Accuracy()

    with pytest.raises(ValueError, match=r"For binary cases, y must be comprised of 0's and 1's"):
        # y has not only 0 or 1 values
        acc.update((torch.randint(0, 2, size=(10,)).long(), torch.arange(0, 10).long()))

    with pytest.raises(ValueError, match=r"For binary cases, y_pred must be comprised of 0's and 1's"):
        # y_pred values are not thresholded to 0, 1 values
        acc.update((torch.rand(10), torch.randint(0, 2, size=(10,)).long()))

    with pytest.raises(ValueError, match=r"y must have shape of "):
        # incompatible shapes
        acc.update((torch.randint(0, 2, size=(10,)).long(), torch.randint(0, 2, size=(10, 5)).long()))

    with pytest.raises(ValueError, match=r"y must have shape of "):
        # incompatible shapes
        acc.update((torch.randint(0, 2, size=(10, 5, 6)).long(), torch.randint(0, 2, size=(10,)).long()))

    with pytest.raises(ValueError, match=r"y must have shape of "):
        # incompatible shapes
        acc.update((torch.randint(0, 2, size=(10,)).long(), torch.randint(0, 2, size=(10, 5, 6)).long()))


@pytest.mark.parametrize("n_times", range(3))
def test_binary_input(n_times, available_device, test_data_binary):
    acc = Accuracy(device=available_device)
    assert acc._device == torch.device(available_device)

    y_pred, y, batch_size = test_data_binary
    acc.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            acc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        acc.update((y_pred, y))

    np_y = y.numpy().ravel()
    np_y_pred = y_pred.numpy().ravel()

    assert acc._type == "binary"
    assert isinstance(acc.compute(), float)
    assert accuracy_score(np_y, np_y_pred) == pytest.approx(acc.compute())


def test_multiclass_wrong_inputs():
    acc = Accuracy()

    with pytest.raises(ValueError):
        # incompatible shapes
        acc.update((torch.rand(10, 5, 4), torch.randint(0, 2, size=(10,)).long()))

    with pytest.raises(ValueError):
        # incompatible shapes
        acc.update((torch.rand(10, 5, 6), torch.randint(0, 5, size=(10, 5)).long()))

    with pytest.raises(ValueError):
        # incompatible shapes
        acc.update((torch.rand(10), torch.randint(0, 5, size=(10, 5, 6)).long()))


@pytest.mark.parametrize("n_times", range(3))
def test_multiclass_input(n_times, available_device, test_data_multiclass):
    acc = Accuracy(device=available_device)
    assert acc._device == torch.device(available_device)

    y_pred, y, batch_size = test_data_multiclass
    acc.reset()
    if batch_size > 1:
        # Batched Updates
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            acc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        acc.update((y_pred, y))

    np_y_pred = y_pred.numpy().argmax(axis=1).ravel()
    np_y = y.numpy().ravel()

    assert acc._type == "multiclass"
    assert isinstance(acc.compute(), float)
    assert accuracy_score(np_y, np_y_pred) == pytest.approx(acc.compute())


def to_numpy_multilabel(y):
    # reshapes input array to (N x ..., C)
    y = y.transpose(1, 0).cpu().numpy()
    num_classes = y.shape[0]
    y = y.reshape((num_classes, -1)).transpose(1, 0)
    return y


def test_multilabel_wrong_inputs():
    acc = Accuracy(is_multilabel=True)

    with pytest.raises(ValueError):
        # incompatible shapes
        acc.update((torch.randint(0, 2, size=(10,)), torch.randint(0, 2, size=(10,)).long()))

    with pytest.raises(ValueError):
        # incompatible y_pred
        acc.update((torch.rand(10, 5), torch.randint(0, 2, size=(10, 5)).long()))

    with pytest.raises(ValueError):
        # incompatible y
        acc.update((torch.randint(0, 5, size=(10, 5, 6)), torch.rand(10)))

    with pytest.raises(ValueError):
        # incompatible binary shapes
        acc.update((torch.randint(0, 2, size=(10, 1)), torch.randint(0, 2, size=(10, 1)).long()))


@pytest.mark.parametrize("n_times", range(3))
def test_multilabel_input(n_times, available_device, test_data_multilabel):
    acc = Accuracy(is_multilabel=True, device=available_device)
    assert acc._device == torch.device(available_device)

    y_pred, y, batch_size = test_data_multilabel
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            acc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        acc.update((y_pred, y))

    np_y_pred = to_numpy_multilabel(y_pred)
    np_y = to_numpy_multilabel(y)

    assert acc._type == "multilabel"
    assert isinstance(acc.compute(), float)
    assert accuracy_score(np_y, np_y_pred) == pytest.approx(acc.compute())


def test_incorrect_type():
    acc = Accuracy()

    # Start as binary data
    y_pred = torch.randint(0, 2, size=(4,))
    y = torch.ones(4).long()
    acc.update((y_pred, y))

    # And add a multiclass data
    y_pred = torch.rand(4, 4)
    y = torch.ones(4).long()

    with pytest.raises(RuntimeError):
        acc.update((y_pred, y))


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_multilabel_input_NHW(self):
        # Multilabel input data of shape (N, C, H, W, ...) and (N, C, H, W, ...)
        rank = idist.get_rank()
        torch.manual_seed(10 + rank)

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            acc = Accuracy(is_multilabel=True, device=metric_device)

            y_pred = torch.randint(0, 2, size=(4, 5, 8, 10), device=device).long()
            y = torch.randint(0, 2, size=(4, 5, 8, 10), device=device).long()
            acc.update((y_pred, y))

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            n = acc._num_examples
            assert n == y.numel() / y.size(dim=1)

            # gather y_pred, y
            y_pred = idist.all_gather(y_pred)
            y = idist.all_gather(y)

            np_y_pred = to_numpy_multilabel(y_pred.cpu())  # (N, C, H, W, ...) -> (N * H * W ..., C)
            np_y = to_numpy_multilabel(y.cpu())  # (N, C, H, W, ...) -> (N * H * W ..., C)
            assert acc._type == "multilabel"
            res = acc.compute()
            assert n == acc._num_examples
            assert isinstance(res, float)
            assert accuracy_score(np_y, np_y_pred) == pytest.approx(res)

            acc.reset()
            torch.manual_seed(10 + rank)
            y_pred = torch.randint(0, 2, size=(4, 7, 10, 8), device=device).long()
            y = torch.randint(0, 2, size=(4, 7, 10, 8), device=device).long()
            acc.update((y_pred, y))

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            n = acc._num_examples
            assert n == y.numel() / y.size(dim=1)

            # gather y_pred, y
            y_pred = idist.all_gather(y_pred)
            y = idist.all_gather(y)

            np_y_pred = to_numpy_multilabel(y_pred.cpu())  # (N, C, H, W, ...) -> (N * H * W ..., C)
            np_y = to_numpy_multilabel(y.cpu())  # (N, C, H, W, ...) -> (N * H * W ..., C)

            assert acc._type == "multilabel"
            res = acc.compute()
            assert n == acc._num_examples
            assert isinstance(res, float)
            assert accuracy_score(np_y, np_y_pred) == pytest.approx(res)
            # check that result is not changed
            res = acc.compute()
            assert n == acc._num_examples
            assert isinstance(res, float)
            assert accuracy_score(np_y, np_y_pred) == pytest.approx(res)

            # Batched Updates
            acc.reset()
            torch.manual_seed(10 + rank)
            y_pred = torch.randint(0, 2, size=(80, 5, 8, 10), device=device).long()
            y = torch.randint(0, 2, size=(80, 5, 8, 10), device=device).long()

            batch_size = 16
            n_iters = y.shape[0] // batch_size + 1

            for i in range(n_iters):
                idx = i * batch_size
                acc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            n = acc._num_examples
            assert n == y.numel() / y.size(dim=1)

            # gather y_pred, y
            y_pred = idist.all_gather(y_pred)
            y = idist.all_gather(y)

            np_y_pred = to_numpy_multilabel(y_pred.cpu())  # (N, C, L, ...) -> (N * L * ..., C)
            np_y = to_numpy_multilabel(y.cpu())  # (N, C, L, ...) -> (N * L ..., C)

            assert acc._type == "multilabel"
            res = acc.compute()
            assert n == acc._num_examples
            assert isinstance(res, float)
            assert accuracy_score(np_y, np_y_pred) == pytest.approx(res)

    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration_multiclass(self, n_epochs):
        rank = idist.get_rank()
        torch.manual_seed(10 + rank)

        n_iters = 80
        batch_size = 16
        n_classes = 10

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            y_true = torch.randint(0, n_classes, size=(n_iters * batch_size,)).to(device)
            y_preds = torch.rand(n_iters * batch_size, n_classes).to(device)

            def update(engine, i):
                return (
                    y_preds[i * batch_size : (i + 1) * batch_size, :],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )

            engine = Engine(update)

            acc = Accuracy(device=metric_device)
            acc.attach(engine, "acc")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_true = idist.all_gather(y_true)
            y_preds = idist.all_gather(y_preds)

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            assert "acc" in engine.state.metrics
            res = engine.state.metrics["acc"]
            if isinstance(res, torch.Tensor):
                res = res.cpu().numpy()

            true_res = accuracy_score(y_true.cpu().numpy(), torch.argmax(y_preds, dim=1).cpu().numpy())

            assert pytest.approx(res) == true_res

            metric_state = acc.state_dict()
            saved__num_correct = acc._num_correct
            saved__num_examples = acc._num_examples
            acc.reset()
            acc.load_state_dict(metric_state)
            assert acc._num_examples == saved__num_examples
            assert (acc._num_correct == saved__num_correct).all()

    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration_multilabel(self, n_epochs):
        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        n_iters = 80
        batch_size = 16
        n_classes = 10

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            metric_device = torch.device(metric_device)

            y_true = torch.randint(0, 2, size=(n_iters * batch_size, n_classes, 8, 10)).to(device)
            y_preds = torch.randint(0, 2, size=(n_iters * batch_size, n_classes, 8, 10)).to(device)

            def update(engine, i):
                return (
                    y_preds[i * batch_size : (i + 1) * batch_size, ...],
                    y_true[i * batch_size : (i + 1) * batch_size, ...],
                )

            engine = Engine(update)

            acc = Accuracy(is_multilabel=True, device=metric_device)
            acc.attach(engine, "acc")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_true = idist.all_gather(y_true)
            y_preds = idist.all_gather(y_preds)

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            assert "acc" in engine.state.metrics
            res = engine.state.metrics["acc"]
            if isinstance(res, torch.Tensor):
                res = res.cpu().numpy()

            true_res = accuracy_score(to_numpy_multilabel(y_true), to_numpy_multilabel(y_preds))

            assert pytest.approx(res) == true_res

    def test_accumulator_device(self):
        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            acc = Accuracy(device=metric_device)
            assert acc._device == metric_device
            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            y_pred = torch.randint(0, 2, size=(10,), device=device, dtype=torch.long)
            y = torch.randint(0, 2, size=(10,), device=device, dtype=torch.long)
            acc.update((y_pred, y))

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration_list_of_tensors_or_numbers(self, n_epochs):
        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        n_iters = 80
        batch_size = 16
        n_classes = 10

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            y_true = torch.randint(0, n_classes, size=(n_iters * batch_size,)).to(device)
            y_preds = torch.rand(n_iters * batch_size, n_classes).to(device)

            def update(_, i):
                return (
                    [v for v in y_preds[i * batch_size : (i + 1) * batch_size, ...]],
                    [v.item() for v in y_true[i * batch_size : (i + 1) * batch_size]],
                )

            engine = Engine(update)

            acc = Accuracy(device=metric_device)
            acc.attach(engine, "acc")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_true = idist.all_gather(y_true)
            y_preds = idist.all_gather(y_preds)

            assert (
                acc._num_correct.device == metric_device
            ), f"{type(acc._num_correct.device)}:{acc._num_correct.device} vs {type(metric_device)}:{metric_device}"

            assert "acc" in engine.state.metrics
            res = engine.state.metrics["acc"]
            if isinstance(res, torch.Tensor):
                res = res.cpu().numpy()

            true_res = accuracy_score(y_true.cpu().numpy(), torch.argmax(y_preds, dim=1).cpu().numpy())
            assert pytest.approx(res) == true_res


def test_skip_unrolling():
    class DummyAcc(Accuracy):
        def __init__(
            self,
            true_output,
            output_transform: Callable = lambda x: x,
            is_multilabel: bool = False,
            device: Union[str, torch.device] = torch.device("cpu"),
            skip_unrolling: bool = False,
        ):
            super(DummyAcc, self).__init__(
                output_transform=output_transform, is_multilabel=False, device=device, skip_unrolling=skip_unrolling
            )
            self.true_output = true_output

        def update(self, output):
            assert output == self.true_output

    a_pred = torch.randint(0, 2, size=(8, 1))
    b_pred = torch.randint(0, 2, size=(8, 1))
    y_pred = [a_pred, b_pred]
    a_true = torch.randint(0, 2, size=(8, 1))
    b_true = torch.randint(0, 2, size=(8, 1))
    y_true = [a_true, b_true]

    acc = DummyAcc(true_output=(y_pred, y_true), skip_unrolling=True)
    state = State(output=(y_pred, y_true))
    engine = MagicMock(state=state)
    acc.iteration_completed(engine)



import logging
import platform
import sys
from collections import namedtuple

import pytest
import torch
from packaging.version import Version

from ignite.engine import Engine, Events
from ignite.utils import _to_str_list, convert_tensor, deprecated, hash_checkpoint, setup_logger, to_onehot


def test_convert_tensor():
    x = torch.tensor([0.0])
    tensor = convert_tensor(x)
    assert torch.is_tensor(tensor)

    x = torch.tensor([0.0])
    tensor = convert_tensor(x, device="cpu", non_blocking=True)
    assert torch.is_tensor(tensor)

    x = torch.tensor([0.0])
    tensor = convert_tensor(x, device="cpu", non_blocking=False)
    assert torch.is_tensor(tensor)

    x = [torch.tensor([0.0]), torch.tensor([0.0])]
    list_ = convert_tensor(x)
    assert isinstance(list_, list)
    assert torch.is_tensor(list_[0])
    assert torch.is_tensor(list_[1])

    x = (torch.tensor([0.0]), torch.tensor([0.0]))
    tuple_ = convert_tensor(x)
    assert isinstance(tuple_, tuple)
    assert torch.is_tensor(tuple_[0])
    assert torch.is_tensor(tuple_[1])

    Point = namedtuple("Point", ["x", "y"])
    x = Point(torch.tensor([0.0]), torch.tensor([0.0]))
    tuple_ = convert_tensor(x)
    assert isinstance(tuple_, Point)
    assert torch.is_tensor(tuple_[0])
    assert torch.is_tensor(tuple_[1])

    x = {"a": torch.tensor([0.0]), "b": torch.tensor([0.0])}
    dict_ = convert_tensor(x)
    assert isinstance(dict_, dict)
    assert torch.is_tensor(dict_["a"])
    assert torch.is_tensor(dict_["b"])

    assert convert_tensor("a") == "a"

    with pytest.raises(TypeError):
        convert_tensor(12345)


@pytest.mark.parametrize(
    "input_data,expected",
    [
        (42, ["42.0000"]),
        ([{"a": 15, "b": torch.tensor([2.0])}], ["a: 15.0000", "b: [2.0000]"]),
        ({"a": 10, "b": 2.33333}, ["a: 10.0000", "b: 2.3333"]),
        ({"x": torch.tensor(0.1234), "y": [1, 2.3567]}, ["x: 0.1234", "y: 1.0000, 2.3567"]),
        (({"nested": [3.1415, torch.tensor(0.0001)]},), ["nested: 3.1415, 0.0001"]),
        (
            {"large_vector": torch.tensor(range(20))},
            ["large_vector: [0.0000, 1.0000, 2.0000, 3.0000, 4.0000, 5.0000, 6.0000, 7.0000, 8.0000, 9.0000, ...]"],
        ),
        ({"large_matrix": torch.randn(5, 5)}, ["large_matrix: Shape[5, 5]"]),
        ({"empty": []}, ["empty: "]),
        ([], []),
        ({"none": None}, ["none: "]),
        ({1: 100, 2: 200}, ["1: 100.0000", "2: 200.0000"]),
    ],
)
def test__to_str_list(input_data, expected):
    assert _to_str_list(input_data) == expected


def test_to_onehot():
    indices = torch.tensor([0, 1, 2, 3], dtype=torch.long)
    actual = to_onehot(indices, 4)
    expected = torch.eye(4, dtype=torch.uint8)
    assert actual.equal(expected)

    y = torch.randint(0, 21, size=(1000,))
    y_ohe = to_onehot(y, num_classes=21)
    y2 = torch.argmax(y_ohe, dim=1)
    assert y.equal(y2)

    y = torch.randint(0, 21, size=(4, 250, 255))
    y_ohe = to_onehot(y, num_classes=21)
    y2 = torch.argmax(y_ohe, dim=1)
    assert y.equal(y2)

    y = torch.randint(0, 21, size=(4, 150, 155, 4, 6))
    y_ohe = to_onehot(y, num_classes=21)
    y2 = torch.argmax(y_ohe, dim=1)
    assert y.equal(y2)

    # Test with `TorchScript`

    x = torch.tensor([0, 1, 2, 3])

    # Test the raw `to_onehot` function
    scripted_to_onehot = torch.jit.script(to_onehot)
    assert scripted_to_onehot(x, 4).allclose(to_onehot(x, 4))

    # Test inside `torch.nn.Module`
    class SLP(torch.nn.Module):
        def __init__(self):
            super(SLP, self).__init__()
            self.linear = torch.nn.Linear(4, 1)

        def forward(self, x):
            x = to_onehot(x, 4)
            return self.linear(x.to(torch.float))

    eager_model = SLP()
    scripted_model = torch.jit.script(eager_model)

    assert eager_model(x).allclose(scripted_model(x))


def test_dist_setup_logger():
    logger = setup_logger("trainer", level=logging.CRITICAL, distributed_rank=1)
    assert logger.level != logging.CRITICAL


def test_setup_logger(capsys, dirname):
    trainer = Engine(lambda e, b: None)
    evaluator = Engine(lambda e, b: None)

    assert len(trainer.logger.handlers) == 0
    trainer.logger.addHandler(logging.NullHandler())
    trainer.logger.addHandler(logging.NullHandler())
    trainer.logger.addHandler(logging.NullHandler())

    fp = dirname / "log"

    def _test(stream):
        trainer.logger = setup_logger("trainer", stream=stream, filepath=fp, reset=True)
        evaluator.logger = setup_logger("evaluator", stream=stream, filepath=fp, reset=True)

        assert len(trainer.logger.handlers) == 2
        assert len(evaluator.logger.handlers) == 2

        @trainer.on(Events.EPOCH_COMPLETED)
        def _(_):
            evaluator.run([0, 1, 2])

        trainer.run([0, 1, 2, 3, 4, 5], max_epochs=5)

        captured = capsys.readouterr()
        if stream is sys.stdout:
            err = captured.out.split("\n")
        else:
            err = captured.err.split("\n")

        with open(fp, "r") as h:
            data = h.readlines()

        for source in [err, data]:
            assert "trainer INFO: Engine run starting with max_epochs=5." in source[0]
            assert "evaluator INFO: Engine run starting with max_epochs=1." in source[1]

    _test(stream=None)
    _test(stream=sys.stderr)
    _test(stream=sys.stdout)

    # Needed by windows to release FileHandler in the loggers
    logging.shutdown()


def _setup_a_logger_and_dump(name, message):
    logger = setup_logger(name)
    logger.info(message)


def test_override_setup_logger(capsys):
    _setup_a_logger_and_dump(__name__, "test_override_setup_logger")

    source = capsys.readouterr().err.split("\n")

    assert "tests.ignite.test_utils INFO: test_override_setup_logger" in source[0]

    # change the logger level of _setup_a_logger_and_dump
    setup_logger(name=__name__, level=logging.WARNING, reset=True)

    _setup_a_logger_and_dump(__name__, "test_override_setup_logger")

    source = capsys.readouterr().err.split("\n")

    assert source[0] == ""

    # Needed by windows to release FileHandler in the loggers
    logging.shutdown()


@pytest.mark.parametrize("encoding", [None, "utf-8"])
def test_setup_logger_encoding(encoding, dirname):
    fp = dirname / "log.txt"
    logger = setup_logger(name="logger", filepath=fp, encoding=encoding, reset=True)
    test_words = ["say hello", "say 你好", "say こんにちわ", "say 안녕하세요", "say привет"]
    for w in test_words:
        logger.info(w)
    logging.shutdown()

    with open(fp, "r", encoding=encoding) as h:
        data = h.readlines()

    if platform.system() == "Windows" and encoding is None:
        flatten_data = "\n".join(data)
        assert test_words[0] in flatten_data
        for word in test_words[1:]:
            assert word not in flatten_data
    else:
        assert len(data) == len(test_words)
        for expected, output in zip(test_words, data):
            assert expected in output


def test_deprecated():
    # Test on function without docs, @deprecated without reasons
    @deprecated("0.4.2", "0.6.0")
    def func_no_docs():
        return 24

    assert func_no_docs.__doc__ == "**Deprecated function**.\n\n    .. deprecated:: 0.4.2"

    # Test on function with docs, @deprecated without reasons
    @deprecated("0.4.2", "0.6.0")
    def func_no_reasons():
        """Docs are cool"""
        return 24

    assert func_no_reasons.__doc__ == "**Deprecated function**.\n\n    Docs are cool.. deprecated:: 0.4.2"

    # Test on function with docs, @deprecated with reasons
    @deprecated("0.4.2", "0.6.0", reasons=("r1", "r2"))
    def func_no_warnings():
        """Docs are very cool"""
        return 24

    assert (
        func_no_warnings.__doc__
        == "**Deprecated function**.\n\n    Docs are very cool.. deprecated:: 0.4.2\n\n\t\n\t- r1\n\t- r2"
    )

    # Tests that the function emits DeprecationWarning
    @deprecated("0.4.2", "0.6.0", reasons=("r1", "r2"))
    def func_check_warning():
        """Docs are very ..."""
        return 24

    with pytest.deprecated_call():
        assert func_check_warning() == 24
    with pytest.warns(
        DeprecationWarning,
        match="This function has been deprecated since version 0.4.2 and will be removed in version 0.6.0."
        + "\n Please refer to the documentation for more details.",
    ):
        # Trigger a warning.
        func_check_warning()

    # Test that the function raises Exception
    @deprecated("0.4.2", "0.6.0", reasons=("reason1", "reason2"), raise_exception=True)
    def func_with_everything():
        return 1

    with pytest.raises(Exception) as exec_info:
        func_with_everything()

    assert (
        str(exec_info.value)
        == "This function has been deprecated since version 0.4.2 and will be removed in version 0.6.0."
        + "\n Please refer to the documentation for more details."
    )


def test_smoke__utils():
    from ignite._utils import apply_to_tensor, apply_to_type, convert_tensor, to_onehot  # noqa: F401


@pytest.mark.skipif(Version(torch.__version__) < Version("1.5.0"), reason="Skip if < 1.5.0")
def test_hash_checkpoint(tmp_path):
    # download lightweight model
    from torchvision.models import squeezenet1_0

    model = squeezenet1_0()
    torch.hub.download_url_to_file(
        "https://download.pytorch.org/models/squeezenet1_0-b66bff10.pth", f"{tmp_path}/squeezenet1_0.pt"
    )
    hash_checkpoint_path, sha_hash = hash_checkpoint(f"{tmp_path}/squeezenet1_0.pt", str(tmp_path))
    model.load_state_dict(torch.load(str(hash_checkpoint_path), "cpu"), True)
    assert sha_hash[:8] == "b66bff10"
    assert hash_checkpoint_path.name == f"squeezenet1_0-{sha_hash[:8]}.pt"

    # test non-existent checkpoint_path
    with pytest.raises(FileNotFoundError, match=r"not_found.pt does not exist in *"):
        hash_checkpoint(f"{tmp_path}/not_found.pt", tmp_path)



from packaging.version import Version
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import torch
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts, ExponentialLR, StepLR

from ignite.engine import Engine, Events
from ignite.handlers.param_scheduler import (
    ConcatScheduler,
    CosineAnnealingScheduler,
    create_lr_scheduler_with_warmup,
    LinearCyclicalScheduler,
    LRScheduler,
    ParamGroupScheduler,
    ParamScheduler,
    PiecewiseLinear,
    ReduceLROnPlateauScheduler,
)
from tests.ignite.handlers import MockFP16DeepSpeedZeroOptimizer

try:
    from torch.optim.lr_scheduler import MultiplicativeLR
except ImportError:
    has_multiplicative_lr = False
else:
    # https://github.com/pytorch/pytorch/issues/32756
    has_multiplicative_lr = Version(torch.__version__) >= Version("1.5.0")


TORCH_GE28 = Version(torch.__version__) >= Version("2.8.0")


class FakeParamScheduler(ParamScheduler):
    def get_param(self):
        return [0]


def test_param_scheduler_asserts():
    t1 = torch.zeros([1], requires_grad=True)
    t2 = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([{"params": t1, "lr": 0.1}, {"params": t2, "lr": 0.1}])

    lr_scheduler = FakeParamScheduler(optimizer, "lr")

    with pytest.raises(ValueError, match=r"size of value is different than optimizer_param_groups"):
        lr_scheduler(None)

    with pytest.raises(TypeError, match=r"Argument state_dict should be a dictionary, but given"):
        lr_scheduler.load_state_dict(None)

    with pytest.raises(ValueError, match=r"Required state attribute 'event_index' is absent in provided state_dict"):
        lr_scheduler.load_state_dict({})

    with pytest.raises(TypeError, match=r"Argument optimizer should be torch.optim.Optimizer"):
        FakeParamScheduler({}, "lr")


def test_linear_scheduler_asserts():
    with pytest.raises(TypeError, match=r"Argument optimizer should be torch.optim.Optimizer"):
        LinearCyclicalScheduler({}, "lr", 1, 0, cycle_size=0)

    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.0)

    with pytest.raises(ValueError, match=r"Argument cycle_size should be positive and larger than 1"):
        LinearCyclicalScheduler(optimizer, "lr", 1, 0, cycle_size=0)

    with pytest.raises(ValueError, match=r"Argument cycle_size should be positive and larger than 1"):
        LinearCyclicalScheduler(optimizer, "lr", 1, 0, cycle_size=1)

    with pytest.raises(
        ValueError,
        match=r"Invalid combination when warmup_duration > 0 and monotonic=False, "
        r"please use either set warmup_duration=0 or monotonic=True",
    ):
        LinearCyclicalScheduler(optimizer, "lr", 1, 0, cycle_size=2, warmup_duration=1)


def test_linear_scheduler():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.0)

    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, 10)
    state_dict = scheduler.state_dict()

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)
    lr_values_in_cycle = [1.0, 0.8, 0.6, 0.4, 0.2, 0.0, 0.2, 0.4, 0.6, 0.8]
    for _ in range(2):
        lrs = []
        trainer.run([0] * 10, max_epochs=2)

        assert lrs == pytest.approx([*lr_values_in_cycle, *lr_values_in_cycle])
        scheduler.load_state_dict(state_dict)

    optimizer = torch.optim.SGD([tensor], lr=0)
    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, 10, cycle_mult=2)
    state_dict = scheduler.state_dict()

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run([0] * 10, max_epochs=3)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    # Cycle 1
                    1.0,
                    0.8,
                    0.6,
                    0.4,
                    0.2,
                    0.0,
                    0.2,
                    0.4,
                    0.6,
                    0.8,
                    # Cycle 2
                    1.0,
                    0.9,
                    0.8,
                    0.7,
                    0.6,
                    0.5,
                    0.4,
                    0.3,
                    0.2,
                    0.1,
                    0.0,
                    0.1,
                    0.2,
                    0.3,
                    0.4,
                    0.5,
                    0.6,
                    0.7,
                    0.8,
                    0.9,
                ],
            )
        )
        scheduler.load_state_dict(state_dict)


def test_linear_scheduler_warmup_duration():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.0)

    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, 10, warmup_duration=5, monotonic=True)
    state_dict = scheduler.state_dict()

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)
    lr_values_in_cycle = [
        1.0,
        0.9,
        0.8,
        0.7,
        0.6,
        0.5,
        0.4,
        0.3,
        0.2,
        0.1,
        0.0,
        0.2,
        0.4,
        0.6,
        0.8,
        1.0,
        0.9,
        0.8,
        0.7,
        0.6,
    ]
    for _ in range(2):
        lrs = []
        trainer.run([0] * 10, max_epochs=2)

        assert lrs == pytest.approx(lr_values_in_cycle)
        scheduler.load_state_dict(state_dict)

    optimizer = torch.optim.SGD([tensor], lr=0)
    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, 10, cycle_mult=2, warmup_duration=5, monotonic=True)
    state_dict = scheduler.state_dict()

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run([0] * 10, max_epochs=3)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    # Cycle 1
                    1.0,
                    0.9,
                    0.8,
                    0.7,
                    0.6,
                    0.5,
                    0.4,
                    0.3,
                    0.2,
                    0.1,
                    0.0,
                    0.2,
                    0.4,
                    0.6,
                    0.8,
                    # Cycle 2
                    1.0,
                    0.95,
                    0.9,
                    0.85,
                    0.8,
                    0.75,
                    0.7,
                    0.65,
                    0.6,
                    0.55,
                    0.5,
                    0.45,
                    0.4,
                    0.35,
                    0.3,
                ],
            )
        )
        scheduler.load_state_dict(state_dict)


def test_linear_scheduler_cycle_size_two():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, cycle_size=2)

    data = [0] * 10
    max_epochs = 2
    simulated_values = LinearCyclicalScheduler.simulate_values(
        num_events=len(data) * max_epochs, param_name="lr", start_value=1, end_value=0, cycle_size=2
    )

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    lrs = []
    trainer.run(data, max_epochs=max_epochs)
    assert lrs == list(
        map(
            pytest.approx,
            [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0],
        )
    )

    assert lrs == pytest.approx([v for i, v in simulated_values])


@pytest.mark.parametrize("cyclic_warmup", [False, True])
def test_cosine_annealing_scheduler(cyclic_warmup):
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler = CosineAnnealingScheduler(optimizer, "lr", 0, 1, 10, warmup_duration=2 if cyclic_warmup else 0)
    state_dict = scheduler.state_dict()

    data = [0] * (10 + int(cyclic_warmup))
    max_epochs = 2
    simulated_values = CosineAnnealingScheduler.simulate_values(
        num_events=len(data) * max_epochs,
        param_name="lr",
        start_value=0,
        end_value=1,
        cycle_size=10,
        warmup_duration=2 if cyclic_warmup else 0,
    )

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)
    lr_values_in_cycle = [
        0.0,
        0.02447174185242318,
        0.09549150281252627,
        0.20610737385376332,
        0.3454915028125263,
        0.5,
        0.6545084971874737,
        0.7938926261462365,
        0.9045084971874737,
        0.9755282581475768,
    ]
    lr_values_in_warmup = np.linspace(1.0, 0.0, 2 + 1)[:-1].tolist() if cyclic_warmup else []

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == pytest.approx([*lr_values_in_cycle, *lr_values_in_warmup, *lr_values_in_cycle])
        scheduler.load_state_dict(state_dict)

        assert lrs == pytest.approx([v for i, v in simulated_values])


def test_concat_scheduler_asserts():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=1.0, end_value=0.0, cycle_size=10)
    scheduler_2 = CosineAnnealingScheduler(optimizer, "lr", start_value=0.0, end_value=1.0, cycle_size=10)

    with pytest.raises(TypeError, match=r"Argument schedulers should be a sequence"):
        ConcatScheduler(schedulers=None, durations=[])

    with pytest.raises(ValueError, match=r"Argument schedulers should be of more than one parameter schedulers"):
        ConcatScheduler(schedulers=[], durations=[])

    with pytest.raises(ValueError, match=r"Argument schedulers should be of more than one parameter schedulers"):
        ConcatScheduler(schedulers=[scheduler_1], durations=[10])

    with pytest.raises(TypeError, match=r"Value at index 1 of schedulers should be a parameter scheduler"):
        ConcatScheduler(schedulers=[scheduler_1, 12], durations=[10])

    with pytest.raises(ValueError, match=r"Incorrect number schedulers or duration values"):
        ConcatScheduler(schedulers=[scheduler_1, scheduler_2], durations=[10, 5])

    with pytest.raises(ValueError, match=r"Argument durations should be list/tuple of integers"):
        ConcatScheduler(schedulers=[scheduler_1, scheduler_2, scheduler_2], durations=[15, 12.0])

    with pytest.raises(TypeError, match=r"Argument durations should be list/tuple"):
        ConcatScheduler(schedulers=[scheduler_1, scheduler_2], durations="abc")

    with pytest.raises(TypeError, match=r"Argument param_names should be list or tuple"):
        ConcatScheduler.simulate_values(
            num_events=123, schedulers=[scheduler_1, scheduler_2], durations=[15], param_names="abc"
        )

    with pytest.raises(ValueError, match=r"Argument param_names should be list or tuple of strings"):
        ConcatScheduler.simulate_values(
            num_events=123, schedulers=[scheduler_1, scheduler_2], durations=[15], param_names=[1]
        )

    optimizer_2 = torch.optim.SGD([tensor], lr=0)
    scheduler_3 = CosineAnnealingScheduler(optimizer_2, "lr", start_value=0.0, end_value=1.0, cycle_size=10)

    with pytest.raises(ValueError, match=r"schedulers should be related to same optimizer"):
        ConcatScheduler([scheduler_1, scheduler_3], durations=[30])

    scheduler_4 = CosineAnnealingScheduler(optimizer, "lr2", start_value=0.0, end_value=1.0, cycle_size=10)

    with pytest.raises(ValueError, match=r"schedulers should be related to same param_name"):
        ConcatScheduler([scheduler_1, scheduler_4], durations=[30])

    with pytest.raises(ValueError, match=r"schedulers should be related to same optimizer"):
        ConcatScheduler.simulate_values(3, [scheduler_1, scheduler_3], durations=[30])


def test_concat_scheduler_state_dict():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)
    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=1.0, end_value=0.0, cycle_size=10)
    scheduler_2 = CosineAnnealingScheduler(optimizer, "lr", start_value=0.0, end_value=1.0, cycle_size=10)
    durations = [10]
    concat_scheduler = ConcatScheduler(schedulers=[scheduler_1, scheduler_2], durations=durations, save_history=False)

    steps = 0
    for i in range(5):
        concat_scheduler(engine=None)
        steps += 1

    state_dict = concat_scheduler.state_dict()

    assert state_dict["durations"] == durations
    assert state_dict["_current_duration"] == durations[0] - steps
    assert state_dict["_scheduler_index"] == 0

    for _ in range(20):
        concat_scheduler(None, None)

    concat_scheduler.load_state_dict(state_dict)
    assert concat_scheduler.durations == durations
    assert concat_scheduler._current_duration == durations[0] - steps
    assert id(concat_scheduler._current_scheduler) == id(scheduler_1)

    with pytest.raises(ValueError, match=r"Required state attribute 'schedulers' is absent in provided state_dict"):
        concat_scheduler.load_state_dict({"a": 1})

    with pytest.raises(ValueError, match=r"Input state_dict contains 0 state_dicts of concatenated schedulers"):
        concat_scheduler.load_state_dict({"schedulers": []})

    with pytest.raises(TypeError, match=r"Argument state_dict should be a dictionary, but given"):
        concat_scheduler.load_state_dict(None)


@pytest.mark.parametrize("duration_vals_as_np_int", [False, True])
def test_concat_scheduler_two_schedulers(duration_vals_as_np_int):
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=1.0, end_value=0.0, cycle_size=10)
    scheduler_2 = CosineAnnealingScheduler(optimizer, "lr", start_value=0.0, end_value=1.0, cycle_size=10)

    durations = [10]
    if duration_vals_as_np_int:
        durations = [np.int64(t) for t in durations]

    concat_scheduler = ConcatScheduler(schedulers=[scheduler_1, scheduler_2], durations=durations, save_history=True)
    state_dict = concat_scheduler.state_dict()

    data = [0] * 10
    max_epochs = 2
    simulated_values = ConcatScheduler.simulate_values(
        num_events=len(data) * max_epochs, schedulers=[scheduler_1, scheduler_2], durations=durations
    )

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, concat_scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    # Cycle 1 of the LinearCyclicalScheduler
                    1.0,
                    0.8,
                    0.6,
                    0.4,
                    0.2,
                    0.0,
                    0.2,
                    0.4,
                    0.6,
                    0.8,
                    # Cycle 1 of the CosineAnnealingScheduler
                    0.0,
                    0.02447174185242318,
                    0.09549150281252627,
                    0.20610737385376332,
                    0.3454915028125263,
                    0.5,
                    0.6545084971874737,
                    0.7938926261462365,
                    0.9045084971874737,
                    0.9755282581475768,
                ],
            )
        )

        state_lrs = trainer.state.param_history["lr"]
        assert len(state_lrs) == len(lrs)
        # Unpack singleton lists
        assert [group[0] for group in state_lrs] == lrs
        assert lrs == pytest.approx([v for i, v in simulated_values])
        concat_scheduler.load_state_dict(state_dict)

        trainer.state.param_history = None


def test_concat_scheduler_two_linear():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=0.0, end_value=0.1, cycle_size=2)
    scheduler_2 = LinearCyclicalScheduler(optimizer, "lr", start_value=0.2, end_value=1.0, cycle_size=2)

    durations = [5]
    concat_scheduler = ConcatScheduler(schedulers=[scheduler_1, scheduler_2], durations=durations, save_history=True)
    state_dict = concat_scheduler.state_dict()

    assert concat_scheduler.get_param() == 0.0

    data = [0] * 10
    max_epochs = 2
    simulated_values = ConcatScheduler.simulate_values(
        num_events=len(data) * max_epochs, schedulers=[scheduler_1, scheduler_2], durations=durations
    )

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, concat_scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    # first LinearCyclicalScheduler
                    0.0,
                    0.1,
                    0.0,
                    0.1,
                    0.0,
                    # second LinearCyclicalScheduler
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                    1.0,
                    0.2,
                ],
            )
        )

        state_lrs = trainer.state.param_history["lr"]
        assert len(state_lrs) == len(lrs)
        # Unpack singleton lists
        assert [group[0] for group in state_lrs] == lrs

        assert lrs == pytest.approx([v for i, v in simulated_values])
        concat_scheduler.load_state_dict(state_dict)

        trainer.state.param_history = None


def test_concat_scheduler_3_schedulers():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=1.0, end_value=0.5, cycle_size=20)
    scheduler_2 = LinearCyclicalScheduler(optimizer, "lr", start_value=0.5, end_value=0.45, cycle_size=10)
    scheduler_3 = LinearCyclicalScheduler(optimizer, "lr", start_value=0.5, end_value=0.0, cycle_size=20)
    durations = [10, 5]

    concat_scheduler = ConcatScheduler(
        schedulers=[scheduler_1, scheduler_2, scheduler_3], durations=durations, save_history=True
    )
    state_dict = concat_scheduler.state_dict()

    data = [0] * 10
    max_epochs = 2
    simulated_values = ConcatScheduler.simulate_values(
        num_events=len(data) * max_epochs, schedulers=[scheduler_1, scheduler_2, scheduler_3], durations=durations
    )

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_STARTED, concat_scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    # Cycle 1 of the first LinearCyclicalScheduler
                    1.0,
                    0.95,
                    0.9,
                    0.85,
                    0.8,
                    0.75,
                    0.7,
                    0.65,
                    0.6,
                    0.55,
                    # Cycle 1 of the second LinearCyclicalScheduler
                    0.5,
                    0.49,
                    0.48,
                    0.47,
                    0.46,
                    # Cycle 1 of the third LinearCyclicalScheduler
                    0.5,
                    0.45,
                    0.4,
                    0.35,
                    0.3,
                ],
            )
        )

        state_lrs = trainer.state.param_history["lr"]
        assert len(state_lrs) == len(lrs)
        # Unpack singleton lists
        assert [group[0] for group in state_lrs] == lrs

        assert lrs == pytest.approx([v for i, v in simulated_values])
        concat_scheduler.load_state_dict(state_dict)

        trainer.state.param_history = None


def test_save_param_history():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    scheduler = LinearCyclicalScheduler(optimizer, "lr", 1, 0, 10, save_history=True)
    lrs = []

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    assert not hasattr(trainer.state, "param_history")

    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)
    trainer.run([0] * 10, max_epochs=2)

    state_lrs = trainer.state.param_history["lr"]
    assert len(state_lrs) == len(lrs)
    # Unpack singleton lists
    assert [group[0] for group in state_lrs] == lrs


def test_lr_scheduler_asserts():
    err_msg = r"Argument lr_scheduler should be a subclass of torch.optim.lr_scheduler.(_LRScheduler|LRScheduler)"
    with pytest.raises(TypeError, match=err_msg):
        LRScheduler(123)

    with pytest.raises(TypeError, match=err_msg):
        LRScheduler.simulate_values(1, None)


@pytest.mark.xfail
@pytest.mark.order(1)
@pytest.mark.parametrize(
    "torch_lr_scheduler_cls, kwargs",
    [
        (ExponentialLR, ({"gamma": 0.78})),
        (MultiplicativeLR if has_multiplicative_lr else None, ({"lr_lambda": lambda epoch: 0.95})),
        (StepLR, ({"step_size": 5, "gamma": 0.5})),
    ],
)
def test_lr_scheduler(torch_lr_scheduler_cls, kwargs):
    if torch_lr_scheduler_cls is None:
        return

    if TORCH_GE28 and torch_lr_scheduler_cls in [ExponentialLR, MultiplicativeLR]:
        pytest.xfail("lr scheduler issues with nightly torch builds")

    tensor = torch.zeros([1], requires_grad=True)
    optimizer1 = torch.optim.SGD([tensor], lr=0.01)
    optimizer2 = torch.optim.SGD([tensor], lr=0.01)
    optimizer3 = torch.optim.SGD([tensor], lr=0.01)
    opt_state_dict1 = optimizer1.state_dict()
    opt_state_dict2 = optimizer2.state_dict()
    opt_state_dict3 = optimizer3.state_dict()

    torch_lr_scheduler1 = torch_lr_scheduler_cls(optimizer=optimizer1, **kwargs)
    scheduler1 = LRScheduler(torch_lr_scheduler1)
    state_dict1 = scheduler1.state_dict()

    torch_lr_scheduler2 = torch_lr_scheduler_cls(optimizer=optimizer2, **kwargs)
    with pytest.warns(UserWarning, match=r"the first lr value from the optimizer, otherwise it will be skipped"):
        scheduler2 = LRScheduler(torch_lr_scheduler2, use_legacy=True)
    state_dict2 = scheduler2.state_dict()

    torch_lr_scheduler3 = torch_lr_scheduler_cls(optimizer=optimizer3, **kwargs)
    state_dict3 = torch_lr_scheduler3.state_dict()

    def dummy_update(engine, batch):
        optimizer1.step()
        optimizer2.step()
        optimizer3.step()

    trainer = Engine(dummy_update)
    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler1)

    @trainer.on(Events.ITERATION_STARTED)
    def save_lr1(engine):
        lrs1.append(optimizer1.param_groups[0]["lr"])

    @trainer.on(Events.ITERATION_STARTED)
    def save_lr2(engine):
        lrs2.append(optimizer2.param_groups[0]["lr"])

    @trainer.on(Events.ITERATION_STARTED)
    def save_true_lr(engine):
        lrs_true.append(optimizer3.param_groups[0]["lr"])

    @trainer.on(Events.ITERATION_COMPLETED)
    def torch_lr_scheduler_step(engine):
        torch_lr_scheduler3.step()

    trainer.add_event_handler(Events.ITERATION_COMPLETED, scheduler2)

    for _ in range(2):
        lrs1 = []
        lrs2 = []
        lrs_true = []
        data = [0] * 10
        max_epochs = 2
        trainer.run(data, max_epochs=max_epochs)
        assert lrs_true == pytest.approx(lrs1), f"{_}: {lrs_true} ({len(lrs_true)}) vs {lrs1} ({len(lrs1)})"
        assert lrs_true == pytest.approx(lrs2), f"{_}: {lrs_true} ({len(lrs_true)}) vs {lrs2} ({len(lrs2)})"
        optimizer1.load_state_dict(opt_state_dict1)
        scheduler1.load_state_dict(state_dict1)
        optimizer2.load_state_dict(opt_state_dict2)
        scheduler2.load_state_dict(state_dict2)
        optimizer3.load_state_dict(opt_state_dict3)
        torch_lr_scheduler3.load_state_dict(state_dict3)

    optimizer4 = torch.optim.SGD([tensor], lr=0.01)
    torch_lr_scheduler4 = torch_lr_scheduler_cls(optimizer=optimizer4, **kwargs)

    simulated_values = LRScheduler.simulate_values(num_events=len(data) * max_epochs, lr_scheduler=torch_lr_scheduler4)
    assert lrs1 == pytest.approx([v for i, v in simulated_values])
    assert lrs2 == pytest.approx([v for i, v in simulated_values])


def test_piecewiselinear_asserts():
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    with pytest.raises(TypeError, match=r"Argument milestones_values should be a list or tuple"):
        PiecewiseLinear(optimizer, "lr", milestones_values=None)

    with pytest.raises(ValueError, match=r"Argument milestones_values should be with at least one value"):
        PiecewiseLinear(optimizer, "lr", milestones_values=[])

    with pytest.raises(ValueError, match=r"Argument milestones_values should be a list of pairs"):
        PiecewiseLinear(optimizer, "lr", milestones_values=[(0.5,)])

    with pytest.raises(ValueError, match=r"Argument milestones_values should be a list of pairs"):
        PiecewiseLinear(optimizer, "lr", milestones_values=[(10, 0.5), (0.6,)])

    with pytest.raises(ValueError, match=r"Milestones should be increasing integers"):
        PiecewiseLinear(optimizer, "lr", milestones_values=[(10, 0.5), (5, 0.6)])

    with pytest.raises(TypeError, match=r"Value of a milestone should be integer"):
        PiecewiseLinear(optimizer, "lr", milestones_values=[(0.5, 1)])


@pytest.mark.parametrize("milestones_as_np_int", [True, False])
def test_piecewiselinear(milestones_as_np_int):
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0)

    milestones_values = [(5, 0.5), (15, 1.0), (25, 0.0), (35, 1.0), (40, 0.5)]
    if milestones_as_np_int:
        milestones_values = [(np.int64(t), v) for t, v in milestones_values]

    scheduler = PiecewiseLinear(optimizer, "lr", milestones_values=milestones_values)
    state_dict = scheduler.state_dict()

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    for _ in range(2):
        lrs = []
        trainer.run([0] * 25, max_epochs=2)

        assert lrs == list(
            map(
                pytest.approx,
                [
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.55,
                    0.6,
                    0.65,
                    0.7,
                    0.75,
                    0.8,
                    0.85,
                    0.9,
                    0.95,
                    1.0,
                    0.9,
                    0.8,
                    0.7,
                    0.6,
                    0.5,
                    0.4,
                    0.3,
                    0.2,
                    0.1,
                    0.0,
                    0.1,
                    0.2,
                    0.3,
                    0.4,
                    0.5,
                    0.6,
                    0.7,
                    0.8,
                    0.9,
                    1.0,
                    0.9,
                    0.8,
                    0.7,
                    0.6,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                    0.5,
                ],
            )
        )
        scheduler.load_state_dict(state_dict)


def test_simulate_and_plot_values():
    import matplotlib

    matplotlib.use("Agg")

    def _test(scheduler_cls, **scheduler_kwargs):
        if scheduler_cls == LRScheduler:
            optimizer = scheduler_kwargs["lr_scheduler"].optimizer
        elif scheduler_cls == ConcatScheduler:
            optimizer = scheduler_kwargs["optimizer"]
            del scheduler_kwargs["optimizer"]
        else:
            tensor = torch.zeros([1], requires_grad=True)
            scheduler_kwargs["optimizer"] = torch.optim.SGD([tensor], lr=0.1)
            optimizer = scheduler_kwargs["optimizer"]

        max_epochs = 2
        data = [0] * 10
        simulated_values = scheduler_cls.simulate_values(num_events=len(data) * max_epochs, **scheduler_kwargs)

        scheduler = scheduler_cls(**scheduler_kwargs)

        lrs = []

        def save_lr(engine):
            lrs.append(optimizer.param_groups[0]["lr"])

        trainer = Engine(lambda engine, batch: None)
        trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)
        trainer.add_event_handler(Events.ITERATION_STARTED, save_lr)
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == pytest.approx([v for i, v in simulated_values])

        # reexecute to check if no internal changes
        # simulated_values = scheduler_cls.simulate_values(num_events=len(data) * max_epochs,
        #                                                  save_history=True,  # this will be removed
        #                                                  **scheduler_kwargs)
        # assert lrs == pytest.approx([v for i, v in simulated_values])

        # launch plot values
        scheduler_cls.plot_values(num_events=len(data) * max_epochs, **scheduler_kwargs)

    # LinearCyclicalScheduler
    _test(LinearCyclicalScheduler, param_name="lr", start_value=1.0, end_value=0.0, cycle_size=10)

    # CosineAnnealingScheduler
    _test(CosineAnnealingScheduler, param_name="lr", start_value=1.0, end_value=0.0, cycle_size=10)

    # LRScheduler
    tensor = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.1)
    torch_lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.5)

    _test(LRScheduler, lr_scheduler=torch_lr_scheduler)

    # ConcatScheduler = [LinearCyclicalScheduler, CosineAnnealingScheduler]
    scheduler_1 = LinearCyclicalScheduler(optimizer, "lr", start_value=1.0, end_value=0.0, cycle_size=20)
    scheduler_2 = CosineAnnealingScheduler(optimizer, "lr", start_value=0.0, end_value=1.0, cycle_size=10)
    durations = [10]
    _test(ConcatScheduler, optimizer=optimizer, schedulers=[scheduler_1, scheduler_2], durations=durations)

    # ConcatScheduler = [LinearCyclicalScheduler, LRScheduler]
    tensor = torch.ones([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.001)
    torch_lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=1.5)
    scheduler_1 = LRScheduler(torch_lr_scheduler)
    scheduler_2 = LinearCyclicalScheduler(optimizer, "lr", start_value=0.1, end_value=0.0, cycle_size=10)
    durations = [10]
    _test(ConcatScheduler, optimizer=optimizer, schedulers=[scheduler_1, scheduler_2], durations=durations)

    # PiecewiseLinear
    tensor = torch.ones([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.001)
    _test(
        PiecewiseLinear,
        optimizer=optimizer,
        param_name="lr",
        milestones_values=[(10, 0.5), (20, 0.45), (21, 0.3), (30, 0.1), (40, 0.1)],
    )

    with pytest.raises(ModuleNotFoundError, match=r"This method requires matplotlib to be installed."):
        with patch.dict("sys.modules", {"matplotlib.pyplot": None}):
            _test(
                PiecewiseLinear,
                optimizer=optimizer,
                param_name="lr",
                milestones_values=[(10, 0.5), (20, 0.45), (21, 0.3), (30, 0.1), (40, 0.1)],
            )


def test_create_lr_scheduler_with_warmup_asserts():
    with pytest.raises(TypeError, match=r"Argument lr_scheduler should be a subclass of"):
        create_lr_scheduler_with_warmup(12, warmup_start_value=0.0, warmup_end_value=0.1, warmup_duration=10)

    t1 = torch.zeros([1], requires_grad=True)
    # A) opt lr != warmup_end_value
    optimizer = torch.optim.SGD([t1], lr=0.2)
    torch_lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.98)

    with pytest.raises(ValueError, match=r"Argument warmup_duration should be at least 2 events"):
        create_lr_scheduler_with_warmup(
            torch_lr_scheduler, warmup_start_value=0.0, warmup_end_value=0.1, warmup_duration=1
        )

    with pytest.raises(TypeError, match=r"Argument warmup_duration should be integer"):
        create_lr_scheduler_with_warmup(
            torch_lr_scheduler, warmup_start_value=0.0, warmup_end_value=0.1, warmup_duration="abc"
        )

    with pytest.raises(TypeError, match=r"Argument output_simulated_values should be a list of None"):
        simulated_values = ()
        create_lr_scheduler_with_warmup(
            torch_lr_scheduler,
            warmup_start_value=0.0,
            warmup_end_value=0.1,
            warmup_duration=10,
            output_simulated_values=simulated_values,
        )


@pytest.mark.parametrize(
    "lr_scheduler_name, warmup_start_value, warmup_end_value, warmup_duration, warmup_end_next_value",
    [
        # A) opt lr != warmup_end_value
        ("ExponentialLR", 0.01, 0.05, 10, 0.2),
        ("ExponentialLR", 0.01, 0.05, 2, 0.2),
        # B) opt lr == warmup_end_value
        ("ExponentialLR", 0.01, 0.2, 10, 0.2 * 0.98),
        ("ExponentialLR", 0.01, 0.2, 2, 0.2 * 0.98),
        # C) lr_scheduler start_value != warmup_end_value
        ("LinearCyclicalScheduler", 0.01, 0.05, 10, 0.8),
        ("LinearCyclicalScheduler", 0.01, 0.05, 2, 0.8),
        # D) lr_scheduler start_value == warmup_end_value
        ("LinearCyclicalScheduler", 0.01, 0.8, 10, 0.8 - (0.8 / 5.0)),
        ("LinearCyclicalScheduler", 0.01, 0.8, 2, 0.8 - (0.8 / 5.0)),
        # E) warmup_end_value is None: fall back to case B)
        ("ExponentialLR", 0.01, None, 10, 0.2 * 0.98),
    ],
)
def test_create_lr_scheduler_with_warmup(
    lr_scheduler_name, warmup_start_value, warmup_end_value, warmup_duration, warmup_end_next_value
):
    t1 = torch.zeros([1], requires_grad=True)

    if lr_scheduler_name == "ExponentialLR":
        optimizer = torch.optim.SGD([t1], lr=0.2)
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.98)
    elif lr_scheduler_name == "LinearCyclicalScheduler":
        optimizer = torch.optim.SGD([t1], lr=0.0)
        lr_scheduler = LinearCyclicalScheduler(
            optimizer=optimizer, param_name="lr", start_value=0.8, end_value=0.0, cycle_size=10
        )
    else:
        raise ValueError(f"Unknown name: {lr_scheduler_name}")

    num_iterations = 10
    max_epochs = 20

    if warmup_end_value is None:
        expected_warmup_end_value = optimizer.param_groups[0]["lr"]
    else:
        expected_warmup_end_value = warmup_end_value

    simulated_values = [None] * (num_iterations * max_epochs)
    scheduler = create_lr_scheduler_with_warmup(
        lr_scheduler,
        warmup_start_value=warmup_start_value,
        warmup_end_value=warmup_end_value,
        warmup_duration=warmup_duration,
        output_simulated_values=simulated_values,
    )

    state_dict = scheduler.state_dict()
    trainer = Engine(lambda engine, batch: None)

    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)

    @trainer.on(Events.ITERATION_STARTED)
    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    data = [0] * num_iterations

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == pytest.approx([v for _, v in simulated_values])

        assert lrs[0] == pytest.approx(warmup_start_value), f"lrs={lrs[: warmup_duration + num_iterations]}"
        assert lrs[warmup_duration - 1] == pytest.approx(
            expected_warmup_end_value
        ), f"lrs={lrs[: warmup_duration + num_iterations]}"
        assert lrs[warmup_duration] == pytest.approx(
            warmup_end_next_value
        ), f"lrs={lrs[: warmup_duration + num_iterations]}"
        scheduler.load_state_dict(state_dict)


@pytest.mark.parametrize("save_history", [False, True])
def test_create_lr_scheduler_with_warmup_on_combined_scheduler(save_history):
    # Test with a complex scheduler
    tensor = torch.ones([1], requires_grad=True)
    optimizer = torch.optim.SGD([tensor], lr=0.001)

    max_epochs = 25
    lr_max_value = 0.4
    num_iterations_per_epoch = 128
    num_iterations = max_epochs * num_iterations_per_epoch
    warmup_duration = 5 * num_iterations_per_epoch
    cooldown_duration = 5 * num_iterations_per_epoch

    scheduler_1 = LinearCyclicalScheduler(
        optimizer,
        "lr",
        start_value=lr_max_value,
        end_value=lr_max_value * 0.9,
        cycle_size=(num_iterations - warmup_duration - cooldown_duration) * 2,
    )

    scheduler_2 = LinearCyclicalScheduler(
        optimizer, "lr", start_value=lr_max_value, end_value=0.0, cycle_size=cooldown_duration * 2
    )

    lr_scheduler = ConcatScheduler(
        schedulers=[scheduler_1, scheduler_2],
        durations=[num_iterations - warmup_duration - cooldown_duration],
        save_history=False,
    )
    lr_values = [None] * num_iterations
    scheduler = create_lr_scheduler_with_warmup(
        lr_scheduler,
        warmup_start_value=0.0,
        warmup_end_value=lr_max_value,
        warmup_duration=warmup_duration,
        save_history=save_history,
        output_simulated_values=lr_values,
    )
    state_dict = scheduler.state_dict()

    trainer = Engine(lambda engine, batch: None)

    @trainer.on(Events.ITERATION_COMPLETED)
    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)

    data = [0] * num_iterations_per_epoch

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)

        assert lrs == pytest.approx([v for i, v in lr_values])

        if save_history:
            param_history = trainer.state.param_history["lr"]
            assert lrs == pytest.approx([v[0] for v in param_history])

            trainer.state.param_history = None

        scheduler.load_state_dict(state_dict)


def test_create_lr_scheduler_with_warmup_with_real_model(dummy_model_factory):
    model = dummy_model_factory(with_grads=False, with_frozen_layer=False)
    init_lr = 0.01
    optimizer = torch.optim.SGD(model.parameters(), lr=init_lr)
    scaled_lr = 0.02
    warmup_duration = 5
    step_size = 2
    gamma = 0.97

    output_simulated_values = [None] * 50

    create_lr_scheduler_with_warmup(
        torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma),
        warmup_start_value=0.0,
        warmup_end_value=scaled_lr,
        warmup_duration=warmup_duration,
        output_simulated_values=output_simulated_values,
    )

    assert output_simulated_values[0] == [0, 0.0]
    assert output_simulated_values[warmup_duration - 1] == [warmup_duration - 1, scaled_lr]
    assert output_simulated_values[warmup_duration] == [warmup_duration, init_lr]
    v = [warmup_duration + step_size, init_lr * gamma]
    assert output_simulated_values[warmup_duration + step_size] == v


def test_param_group_scheduler_asserts():
    t1 = torch.zeros([1], requires_grad=True)
    t2 = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([{"params": t1, "lr": 0.1}, {"params": t2, "lr": 0.1}])

    lr_scheduler1 = LinearCyclicalScheduler(
        optimizer, "lr", param_group_index=0, start_value=1.0, end_value=0.0, cycle_size=10
    )
    lr_scheduler2 = LinearCyclicalScheduler(
        optimizer, "lr", param_group_index=1, start_value=1.0, end_value=0.0, cycle_size=10
    )

    with pytest.raises(TypeError, match=r"Argument schedulers should be a list/tuple"):
        ParamGroupScheduler(schedulers=None, names=["a", "b", "c"])

    with pytest.raises(ValueError, match=r"Argument schedulers should be a list/tuple of parameter schedulers"):
        ParamGroupScheduler(schedulers=[0, 1, 2], names=["a", "b", "c"])

    with pytest.raises(ValueError, match=r"Argument schedulers should be a list/tuple of parameter schedulers"):
        ParamGroupScheduler(schedulers=[lr_scheduler1, "2"], names=["a", "b"])

    with pytest.raises(TypeError, match=r"Argument names should be a list/tuple"):
        ParamGroupScheduler(schedulers=[lr_scheduler1, lr_scheduler2], names="ab")

    with pytest.raises(ValueError, match=r"Argument names should be a list/tuple of parameter scheduler's names"):
        ParamGroupScheduler(schedulers=[lr_scheduler1, lr_scheduler2], names=[1, 2])

    with pytest.raises(ValueError, match=r"\d should be equal \d"):
        ParamGroupScheduler(schedulers=[lr_scheduler1, lr_scheduler2], names=["a"])

    scheduler = ParamGroupScheduler(schedulers=[lr_scheduler1, lr_scheduler2], names=["a", "b"])
    with pytest.raises(TypeError, match=r"Argument state_dict should be a dictionary"):
        scheduler.load_state_dict(None)

    with pytest.raises(ValueError, match=r"Required state attribute 'schedulers' is absent in provided state_dict"):
        scheduler.load_state_dict({"a": 1})

    with pytest.raises(ValueError, match=r"Input state_dict contains 0 state_dicts of param group schedulers"):
        scheduler.load_state_dict({"schedulers": []})

    with pytest.raises(ValueError, match=r"Required state attribute 'schedulers' is absent in provided state_dict"):
        scheduler.load_state_dict({})

    with pytest.raises(
        ValueError, match=r"Name of scheduler from input state dict does not " r"correspond to required one"
    ):
        scheduler.load_state_dict({"schedulers": [("a", lr_scheduler1.state_dict()), ("bad_name", {})]})


@pytest.mark.parametrize("param_groups_setting", ["single_optim", "multi_optim"])
def test_param_group_scheduler(param_groups_setting):
    t1 = torch.zeros([1], requires_grad=True)
    t2 = torch.zeros([1], requires_grad=True)
    if param_groups_setting == "single_optim":
        optimizer = torch.optim.SGD([{"params": t1, "lr": 0.1}, {"params": t2, "lr": 0.1}])

        lr_scheduler1 = LinearCyclicalScheduler(
            optimizer, "lr", param_group_index=0, start_value=1.0, end_value=0.0, cycle_size=10
        )
        lr_scheduler2 = LinearCyclicalScheduler(
            optimizer, "lr", param_group_index=1, start_value=1.0, end_value=0.0, cycle_size=10
        )

    else:
        optimizer_1 = torch.optim.SGD(params=[t1], lr=0.1)
        optimizer_2 = torch.optim.SGD(params=[t2], lr=0.1)

        lr_scheduler1 = LinearCyclicalScheduler(optimizer_1, "lr", start_value=1.0, end_value=0.0, cycle_size=10)
        lr_scheduler2 = LinearCyclicalScheduler(optimizer_2, "lr", start_value=1.0, end_value=0.0, cycle_size=10)

    lr_schedulers = [lr_scheduler1, lr_scheduler2]
    num_iterations = 10
    max_epochs = 20

    scheduler = ParamGroupScheduler(lr_schedulers, names=[f"s_{i}" for i in range(len(lr_schedulers))])
    state_dict = scheduler.state_dict()

    trainer = Engine(lambda engine, batch: None)

    lrs = []

    @trainer.on(Events.ITERATION_STARTED, lrs)
    def save_lr(_, lrs):
        lrs.append(scheduler.get_param())

    trainer.add_event_handler(Events.ITERATION_STARTED, scheduler)

    data = [0] * num_iterations

    for _ in range(2):
        lrs.clear()
        trainer.run(data, max_epochs=max_epochs)

        assert [lr[0] for lr in lrs] == pytest.approx([lr[1] for lr in lrs])
        scheduler.load_state_dict(state_dict)

        values = ParamGroupScheduler.simulate_values(max_epochs * num_iterations, lr_schedulers)
        assert [lr[1] for lr in values] == pytest.approx([lr[2] for lr in values])
        assert [lr[0] for lr in lrs] == pytest.approx([lr[1] for lr in values])


@pytest.mark.parametrize(
    "scheduler_cls, kwargs",
    [
        (LinearCyclicalScheduler, {"param_name": "lr", "start_value": 1.0, "end_value": 0.0, "cycle_size": 10}),
        (
            PiecewiseLinear,
            {"param_name": "lr", "milestones_values": [(5, 0.5), (15, 1.0), (25, 0.0), (35, 1.0), (40, 0.5)]},
        ),
        (CosineAnnealingScheduler, {"param_name": "lr", "start_value": 0.0, "end_value": 1.0, "cycle_size": 10}),
        (ExponentialLR, {"gamma": 0.98}),
        (StepLR, {"step_size": 50, "gamma": 0.5}),
    ],
)
def test_scheduler_with_param_groups(scheduler_cls, kwargs):
    t1 = torch.zeros([1], requires_grad=True)
    t2 = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([{"params": t1, "lr": 0.1}, {"params": t2, "lr": 0.1}])

    lr_scheduler = scheduler_cls(optimizer, **kwargs)
    if not isinstance(lr_scheduler, ParamScheduler):
        lr_scheduler = LRScheduler(lr_scheduler)

    num_iterations = 10
    max_epochs = 20

    state_dict = lr_scheduler.state_dict()

    trainer = Engine(lambda engine, batch: None)

    @trainer.on(Events.ITERATION_COMPLETED)
    def save_lr():
        lrs.append((optimizer.param_groups[0]["lr"], optimizer.param_groups[1]["lr"]))

    trainer.add_event_handler(Events.ITERATION_STARTED, lr_scheduler)

    data = [0] * num_iterations

    for _ in range(2):
        lrs = []
        trainer.run(data, max_epochs=max_epochs)
        assert [lr[0] for lr in lrs] == pytest.approx([lr[1] for lr in lrs])
        lr_scheduler.load_state_dict(state_dict)


def test_lr_scheduling_on_non_torch_optimizers():
    # tests https://github.com/pytorch/ignite/issues/1162
    optimizer = MagicMock()
    optimizer.param_groups = [{"params": 0}]
    FakeParamScheduler(optimizer, "lr")

    tensor = torch.zeros([1], requires_grad=True)
    base_optimizer = torch.optim.SGD([tensor], lr=0)
    optimizer = MockFP16DeepSpeedZeroOptimizer(base_optimizer)

    milestones_values = [(5, 0.5), (15, 1.0)]

    scheduler = PiecewiseLinear(optimizer, "lr", milestones_values=milestones_values)

    def save_lr(engine):
        lrs.append(optimizer.param_groups[0]["lr"])

    trainer = Engine(lambda engine, batch: None)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, scheduler)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, save_lr)

    lrs = []
    trainer.run([0] * 15, max_epochs=1)

    assert lrs == list(
        map(pytest.approx, [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95])
    )


def test_reduce_lr_on_plateau_scheduler():
    tensor1 = torch.zeros([1], requires_grad=True)
    tensor2 = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([{"params": [tensor1]}, {"params": [tensor2]}], lr=1)

    data = [0] * 8
    max_epochs = 10

    trainer = Engine(lambda engine, batch: None)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluate():
        evaluator.run(data)

    scheduler = ReduceLROnPlateauScheduler(
        optimizer,
        metric_name="acc",
        mode="max",
        factor=0.5,
        patience=1,
        threshold_mode="abs",
        threshold=1.99,
        min_lr=1e-7,
        save_history=True,
        trainer=trainer,
        param_group_index=0,
    )
    evaluator = Engine(lambda engine, batch: None)
    evaluator.state.metrics = {"acc": 0.0}
    generate_acc = iter([3, 7, 7, 9, 10, 11, 8, 8, 4, 7])

    @evaluator.on(Events.COMPLETED)
    def set_acc():
        evaluator.state.metrics["acc"] = next(generate_acc)

    evaluator.add_event_handler(Events.COMPLETED, scheduler)

    trainer.run(data, max_epochs=max_epochs)

    lrs = [param[0] for param in trainer.state.param_history["lr"]]
    assert lrs == list(
        map(
            pytest.approx,
            [1, 1, 1, 1, 1, 1, 1, 0.5, 0.5, 0.25],
        )
    )
    assert optimizer.param_groups[1]["lr"] == 1

    values = ReduceLROnPlateauScheduler.simulate_values(
        5, [10, 9, 9, 9, 8.1], 1.0, save_history=True, factor=0.5, patience=2, threshold=0.1
    )
    values = np.array(values)[:, 1].tolist()
    assert values == list(
        map(
            pytest.approx,
            [1.0, 1.0, 1.0, 0.5, 0.5],
        )
    )


def test_reduce_lr_on_plateau_scheduler_asserts():
    tensor1 = torch.zeros([1], requires_grad=True)
    tensor2 = torch.zeros([1], requires_grad=True)
    optimizer = torch.optim.SGD([{"params": [tensor1]}, {"params": [tensor2]}], lr=1)

    with pytest.raises(TypeError, match=r"When param_group_index is given, min_lr should be a float, but given"):
        ReduceLROnPlateauScheduler(
            optimizer,
            metric_name="acc",
            min_lr=[1e-7, 1e-8],
            param_group_index=0,
        )

    with pytest.raises(
        ValueError, match=r"Argument engine should have in its 'state', attribute 'metrics' which itself has the metric"
    ):
        scheduler = ReduceLROnPlateauScheduler(optimizer, metric_name="acc")
        evaluator = Engine(lambda engine, batch: None)
        scheduler(evaluator)

    with pytest.raises(ValueError, match=r"Length of argument metric_values should be equal to num_events."):
        metric_values = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
        ReduceLROnPlateauScheduler.simulate_values(5, metric_values, 0.01)


@pytest.mark.parametrize("warmup_end_value", [0.23, None])
@pytest.mark.parametrize("T_0", [1, 12])
@pytest.mark.parametrize("T_mult", [1, 3])
def test_create_lr_scheduler_with_warmup_cosine(warmup_end_value, T_0, T_mult):
    lr = 0.2
    steps = 200
    warm_steps = 50
    warm_start = 0.023

    def get_optim():
        t1 = torch.zeros([1], requires_grad=True)
        return torch.optim.SGD([t1], lr=lr)

    def get_cos_shed():
        return CosineAnnealingWarmRestarts(optimizer, T_0=T_0, T_mult=T_mult)

    optimizer = get_optim()
    scheduler = get_cos_shed()
    cosine_lrs = []
    for i in range(steps):
        cosine_lrs.append(optimizer.param_groups[0]["lr"])
        scheduler.step()

    optimizer = get_optim()
    scheduler = create_lr_scheduler_with_warmup(
        get_cos_shed(), warmup_start_value=warm_start, warmup_end_value=warmup_end_value, warmup_duration=warm_steps
    )

    warm_lrs = []
    real_warm_steps = warm_steps if warmup_end_value is not None else (warm_steps - 1)
    for epoch in range(real_warm_steps + steps):
        scheduler(None)
        warm_lrs.append(optimizer.param_groups[0]["lr"])

    if warmup_end_value is not None:
        np.testing.assert_allclose(np.linspace(warm_start, warmup_end_value, warm_steps), warm_lrs[:warm_steps])
        assert warm_lrs[real_warm_steps:] == cosine_lrs
    else:
        np.testing.assert_allclose(np.linspace(warm_start, lr, warm_steps), warm_lrs[:warm_steps])
        assert warm_lrs[real_warm_steps:] == cosine_lrs
